import json
import numpy as np
import paho.mqtt.client as mqtt
from naeural_core import Logger
from time import sleep

DEVICE_ID = '1234'

_CONFIG_SENDER = {
  'MQTT_HOST': 'HOSTNAME',
  'MQTT_PORT': 1883,
  'MQTT_USER': 'USERNAME',
  'MQTT_PASS': 'PASSWORD',
  'MQTT_PATH': 'PATH',
  'MQTT_TOPIC': 'config1',
  'MQTT_QOS': 1
  }

def on_connect(client, userdata, flags, rc):
  print('Connected with result code {}'.format(rc))
  return

def on_disconnect(client, userdata, rc):
  # if rc != 0:
  print('Disconnected with result code {}'.format(rc))
  return

def on_message(client, userdata, message):
  print('Received message "{}" on topic {} with QoS {}'.format(message, message.topic, str(message.qos)))
  return

def on_publish(client, userdata, mid):
  print('Publish {}, {}, {}'.format(client, userdata, mid))
  return
  
def on_log(mqttc, userdata, level, string):
  print(string)
  return
  

if __name__ == '__main__':  
  cfg_file = 'main_config.txt'
  log = Logger(
    lib_name='TST', 
    config_file=cfg_file, 
    max_lines=1000, 
    TF_KERAS=False
    )
  mqttc = mqtt.Client(
    client_id=DEVICE_ID,
    clean_session=False
    )  
  
  mqttc.username_pw_set(
    username=_CONFIG_SENDER['MQTT_USER'],
    password=_CONFIG_SENDER['MQTT_PASS']
    )
  
  mqttc.on_connect = on_connect
  mqttc.on_disconnect = on_disconnect
  mqttc.on_message = on_message
  mqttc.on_publish = on_publish
  mqttc.on_log = on_log
  
  mqttc.connect(
    host=_CONFIG_SENDER['MQTT_HOST'],
    port=_CONFIG_SENDER['MQTT_PORT']
    )
  
  # mqttc.loop_misc() #nu afiseaza nimic
  mqttc.loop_read()

  # i = 0
  # topic_name = '{}/{}'.format(_CONFIG_SENDER['MQTT_PATH'], _CONFIG_SENDER['MQTT_TOPIC'])
  # while True:
  #   i+= 1
  #   mqttc.publish(topic_name, 'Iter {}'.format(i))
  #   # mqttc.loop(timeout=0.1)
  #   sleep(120)
  
  # while True:
  # for i in range(100):
  #   img = np.random.randint(0, 256, (2000, 3000, 3), dtype=np.uint8)
  #   img_base64 = log.np_image_to_base64(img)
  #   # mqttc.publish(topic_name, 'Iter {}'.format(i), qos=_CONFIG_SENDER['MQTT_QOS'])    
  #   payload = {
  #     'ID': i + 1,
  #     'IMG': img_base64
  #     }
  #   mqttc.publish(
  #     topic=topic_name, 
  #     payload=json.dumps(payload), 
  #     qos=_CONFIG_SENDER['MQTT_QOS']
  #     )    
  #   mqttc.loop(timeout=0.01)
  #   sleep(0.1)
    
    





















# YOLOv5 🚀 by Ultralytics, GPL-3.0 license
"""
YOLO-specific modules

Usage:
    $ python path/to/models/yolo.py --cfg yolov5s.yaml
"""
import logging
import torch
import torch.nn as nn
import math
import argparse
import sys
from copy import deepcopy
from pathlib import Path


from models.common import (
  Conv, GhostConv, Bottleneck, GhostBottleneck, SPP, SPPF, DWConv, Focus, 
  BottleneckCSP, C3, C3TR, C3SPP, C3Ghost, Concat, Contract, Expand,
  AutoShape
  )
from models.experimental import (
  MixConv2d, CrossConv
  )
from y5utils.autoanchor import check_anchor_order
from y5utils.general import check_yaml, make_divisible, print_args, set_logging
from y5utils.plots import feature_visualization
from y5utils.torch_utils import copy_attr, fuse_conv_and_bn, initialize_weights, model_info, scale_img, \
    select_device, time_sync

try:
    import thop  # for FLOPs computation
except ImportError:
    thop = None

LOGGER = logging.getLogger(__name__)


class Detect(nn.Module):
    stride = None  # strides computed during build
    onnx_dynamic = False  # ONNX export parameter

    def __init__(self, nc=80, anchors=(), ch=(), inplace=True):  # detection layer
        super().__init__()
        self.nc = nc  # number of classes
        self.no = nc + 5  # number of outputs per anchor
        self.nl = len(anchors)  # number of detection layers
        self.na = len(anchors[0]) // 2  # number of anchors
        self.grid = [torch.zeros(1)] * self.nl  # init grid
        a = torch.tensor(anchors).float().view(self.nl, -1, 2)
        self.register_buffer('anchors', a)  # shape(nl,na,2)
        self.register_buffer('anchor_grid', a.clone().view(self.nl, 1, -1, 1, 1, 2))  # shape(nl,1,na,1,1,2)
        self.m = nn.ModuleList(nn.Conv2d(x, self.no * self.na, 1) for x in ch)  # output conv
        self.inplace = inplace  # use in-place ops (e.g. slice assignment)

    def forward(self, x):
        z = []  # inference output
        for i in range(self.nl):
            x[i] = self.m[i](x[i])  # conv
            bs, _, ny, nx = x[i].shape  # x(bs,255,20,20) to x(bs,3,20,20,85)
            x[i] = x[i].view(bs, self.na, self.no, ny, nx).permute(0, 1, 3, 4, 2).contiguous()

            if not self.training:  # inference
                if self.grid[i].shape[2:4] != x[i].shape[2:4] or self.onnx_dynamic:
                    self.grid[i] = self._make_grid(nx, ny).to(x[i].device)

                y = x[i].sigmoid()
                if self.inplace:
                    y[..., 0:2] = (y[..., 0:2] * 2. - 0.5 + self.grid[i]) * self.stride[i]  # xy
                    y[..., 2:4] = (y[..., 2:4] * 2) ** 2 * self.anchor_grid[i]  # wh
                else:  # for YOLOv5 on AWS Inferentia https://github.com/ultralytics/yolov5/pull/2953
                    xy = (y[..., 0:2] * 2. - 0.5 + self.grid[i]) * self.stride[i]  # xy
                    wh = (y[..., 2:4] * 2) ** 2 * self.anchor_grid[i].view(1, self.na, 1, 1, 2)  # wh
                    y = torch.cat((xy, wh, y[..., 4:]), -1)
                z.append(y.view(bs, -1, self.no))

        return x if self.training else (torch.cat(z, 1), x)

    @staticmethod
    def _make_grid(nx=20, ny=20):
        yv, xv = torch.meshgrid([torch.arange(ny), torch.arange(nx)])
        return torch.stack((xv, yv), 2).view((1, 1, ny, nx, 2)).float()


class Model(nn.Module):
    def __init__(self, cfg='yolov5s.yaml', ch=3, nc=None, anchors=None):  # model, input channels, number of classes
        super().__init__()
        if isinstance(cfg, dict):
            self.yaml = cfg  # model dict
        else:  # is *.yaml
            import yaml  # for torch hub
            self.yaml_file = Path(cfg).name
            with open(cfg) as f:
                self.yaml = yaml.safe_load(f)  # model dict

        # Define model
        ch = self.yaml['ch'] = self.yaml.get('ch', ch)  # input channels
        if nc and nc != self.yaml['nc']:
            LOGGER.info(f"Overriding model.yaml nc={self.yaml['nc']} with nc={nc}")
            self.yaml['nc'] = nc  # override yaml value
        if anchors:
            LOGGER.info(f'Overriding model.yaml anchors with anchors={anchors}')
            self.yaml['anchors'] = round(anchors)  # override yaml value
        self.model, self.save = parse_model(deepcopy(self.yaml), ch=[ch])  # model, savelist
        self.names = [str(i) for i in range(self.yaml['nc'])]  # default names
        self.inplace = self.yaml.get('inplace', True)

        # Build strides, anchors
        m = self.model[-1]  # Detect()
        if isinstance(m, Detect):
            s = 256  # 2x min stride
            m.inplace = self.inplace
            m.stride = torch.tensor([s / x.shape[-2] for x in self.forward(torch.zeros(1, ch, s, s))])  # forward
            m.anchors /= m.stride.view(-1, 1, 1)
            check_anchor_order(m)
            self.stride = m.stride
            self._initialize_biases()  # only run once

        # Init weights, biases
        initialize_weights(self)
        self.info()
        LOGGER.info('')

    def forward(self, x, augment=False, profile=False, visualize=False):
        if augment:
            return self._forward_augment(x)  # augmented inference, None
        return self._forward_once(x, profile, visualize)  # single-scale inference, train

    def _forward_augment(self, x):
        img_size = x.shape[-2:]  # height, width
        s = [1, 0.83, 0.67]  # scales
        f = [None, 3, None]  # flips (2-ud, 3-lr)
        y = []  # outputs
        for si, fi in zip(s, f):
            xi = scale_img(x.flip(fi) if fi else x, si, gs=int(self.stride.max()))
            yi = self._forward_once(xi)[0]  # forward
            # cv2.imwrite(f'img_{si}.jpg', 255 * xi[0].cpu().numpy().transpose((1, 2, 0))[:, :, ::-1])  # save
            yi = self._descale_pred(yi, fi, si, img_size)
            y.append(yi)
        return torch.cat(y, 1), None  # augmented inference, train

    def _forward_once(self, x, profile=False, visualize=False):
        y, dt = [], []  # outputs
        for m in self.model:
            if m.f != -1:  # if not from previous layer
                x = y[m.f] if isinstance(m.f, int) else [x if j == -1 else y[j] for j in m.f]  # from earlier layers
            if profile:
                self._profile_one_layer(m, x, dt)
            x = m(x)  # run
            y.append(x if m.i in self.save else None)  # save output
            if visualize:
                feature_visualization(x, m.type, m.i, save_dir=visualize)
        return x

    def _descale_pred(self, p, flips, scale, img_size):
        # de-scale predictions following augmented inference (inverse operation)
        if self.inplace:
            p[..., :4] /= scale  # de-scale
            if flips == 2:
                p[..., 1] = img_size[0] - p[..., 1]  # de-flip ud
            elif flips == 3:
                p[..., 0] = img_size[1] - p[..., 0]  # de-flip lr
        else:
            x, y, wh = p[..., 0:1] / scale, p[..., 1:2] / scale, p[..., 2:4] / scale  # de-scale
            if flips == 2:
                y = img_size[0] - y  # de-flip ud
            elif flips == 3:
                x = img_size[1] - x  # de-flip lr
            p = torch.cat((x, y, wh, p[..., 4:]), -1)
        return p

    def _profile_one_layer(self, m, x, dt):
        c = isinstance(m, Detect)  # is final layer, copy input as inplace fix
        o = thop.profile(m, inputs=(x.copy() if c else x,), verbose=False)[0] / 1E9 * 2 if thop else 0  # FLOPs
        t = time_sync()
        for _ in range(10):
            m(x.copy() if c else x)
        dt.append((time_sync() - t) * 100)
        if m == self.model[0]:
            LOGGER.info(f"{'time (ms)':>10s} {'GFLOPs':>10s} {'params':>10s}  {'module'}")
        LOGGER.info(f'{dt[-1]:10.2f} {o:10.2f} {m.np:10.0f}  {m.type}')
        if c:
            LOGGER.info(f"{sum(dt):10.2f} {'-':>10s} {'-':>10s}  Total")

    def _initialize_biases(self, cf=None):  # initialize biases into Detect(), cf is class frequency
        # https://arxiv.org/abs/1708.02002 section 3.3
        # cf = torch.bincount(torch.tensor(np.concatenate(dataset.labels, 0)[:, 0]).long(), minlength=nc) + 1.
        m = self.model[-1]  # Detect() module
        for mi, s in zip(m.m, m.stride):  # from
            b = mi.bias.view(m.na, -1)  # conv.bias(255) to (3,85)
            b.data[:, 4] += math.log(8 / (640 / s) ** 2)  # obj (8 objects per 640 image)
            b.data[:, 5:] += math.log(0.6 / (m.nc - 0.99)) if cf is None else torch.log(cf / cf.sum())  # cls
            mi.bias = torch.nn.Parameter(b.view(-1), requires_grad=True)

    def _print_biases(self):
        m = self.model[-1]  # Detect() module
        for mi in m.m:  # from
            b = mi.bias.detach().view(m.na, -1).T  # conv.bias(255) to (3,85)
            LOGGER.info(
                ('%6g Conv2d.bias:' + '%10.3g' * 6) % (mi.weight.shape[1], *b[:5].mean(1).tolist(), b[5:].mean()))

    # def _print_weights(self):
    #     for m in self.model.modules():
    #         if type(m) is Bottleneck:
    #             LOGGER.info('%10.3g' % (m.w.detach().sigmoid() * 2))  # shortcut weights

    def fuse(self):  # fuse model Conv2d() + BatchNorm2d() layers
        LOGGER.info('Fusing layers... ')
        for m in self.model.modules():
            if isinstance(m, (Conv, DWConv)) and hasattr(m, 'bn'):
                m.conv = fuse_conv_and_bn(m.conv, m.bn)  # update conv
                delattr(m, 'bn')  # remove batchnorm
                m.forward = m.forward_fuse  # update forward
        self.info()
        return self

    def autoshape(self):  # add AutoShape module
        LOGGER.info('Adding AutoShape... ')
        m = AutoShape(self)  # wrap model
        copy_attr(m, self, include=('yaml', 'nc', 'hyp', 'names', 'stride'), exclude=())  # copy attributes
        return m

    def info(self, verbose=False, img_size=640):  # print model information
        model_info(self, verbose, img_size)


def parse_model(d, ch):  # model_dict, input_channels(3)
    LOGGER.info('\n%3s%18s%3s%10s  %-40s%-30s' % ('', 'from', 'n', 'params', 'module', 'arguments'))
    anchors, nc, gd, gw = d['anchors'], d['nc'], d['depth_multiple'], d['width_multiple']
    na = (len(anchors[0]) // 2) if isinstance(anchors, list) else anchors  # number of anchors
    no = na * (nc + 5)  # number of outputs = anchors * (classes + 5)

    layers, save, c2 = [], [], ch[-1]  # layers, savelist, ch out
    for i, (f, n, m, args) in enumerate(d['backbone'] + d['head']):  # from, number, module, args
        m = eval(m) if isinstance(m, str) else m  # eval strings
        for j, a in enumerate(args):
            try:
                args[j] = eval(a) if isinstance(a, str) else a  # eval strings
            except:
                pass

        n = n_ = max(round(n * gd), 1) if n > 1 else n  # depth gain
        if m in [Conv, GhostConv, Bottleneck, GhostBottleneck, SPP, SPPF, DWConv, MixConv2d, Focus, CrossConv,
                 BottleneckCSP, C3, C3TR, C3SPP, C3Ghost]:
            c1, c2 = ch[f], args[0]
            if c2 != no:  # if not output
                c2 = make_divisible(c2 * gw, 8)

            args = [c1, c2, *args[1:]]
            if m in [BottleneckCSP, C3, C3TR, C3Ghost]:
                args.insert(2, n)  # number of repeats
                n = 1
        elif m is nn.BatchNorm2d:
            args = [ch[f]]
        elif m is Concat:
            c2 = sum([ch[x] for x in f])
        elif m is Detect:
            args.append([ch[x] for x in f])
            if isinstance(args[1], int):  # number of anchors
                args[1] = [list(range(args[1] * 2))] * len(f)
        elif m is Contract:
            c2 = ch[f] * args[0] ** 2
        elif m is Expand:
            c2 = ch[f] // args[0] ** 2
        else:
            c2 = ch[f]

        m_ = nn.Sequential(*[m(*args) for _ in range(n)]) if n > 1 else m(*args)  # module
        t = str(m)[8:-2].replace('__main__.', '')  # module type
        np = sum([x.numel() for x in m_.parameters()])  # number params
        m_.i, m_.f, m_.type, m_.np = i, f, t, np  # attach index, 'from' index, type, number params
        LOGGER.info('%3s%18s%3s%10.0f  %-40s%-30s' % (i, f, n_, np, t, args))  # print
        save.extend(x % i for x in ([f] if isinstance(f, int) else f) if x != -1)  # append to savelist
        layers.append(m_)
        if i == 0:
            ch = []
        ch.append(c2)
    return nn.Sequential(*layers), sorted(save)


if __name__ == '__main__':
  import torch as th
  parser = argparse.ArgumentParser()
  parser.add_argument('--cfg', type=str, default='yolov5l6.yaml', help='model.yaml')
  parser.add_argument('--device', default='', help='cuda device, i.e. 0 or 0,1,2,3 or cpu')
  parser.add_argument('--profile', action='store_true', help='profile model speed')
  opt = parser.parse_args()
  opt.cfg = check_yaml(opt.cfg)  # check YAML
  # print_args(FILE.stem, opt)
  set_logging()
  device = select_device(opt.device)

  # Create model
  model = Model(opt.cfg).to(device)
  # model.train()
  
  #TODO: lets load weights & run deterministic test
  fn = '_cache/_models/yolov5l6_weights.pt'
  model.load_state_dict(th.load(fn))
  
  
  
  

  # # Profile
  # if opt.profile:
  #     img = torch.rand(8 if torch.cuda.is_available() else 1, 3, 640, 640).to(device)
  #     y = model(img, profile=True)

  # Tensorboard (not working https://github.com/ultralytics/yolov5/issues/2898)
  # from torch.utils.tensorboard import SummaryWriter
  # tb_writer = SummaryWriter('.')
  # LOGGER.info("Run 'tensorboard --logdir=models' to view tensorboard at http://localhost:6006/")
  # tb_writer.add_graph(torch.jit.trace(model, img, strict=False), [])  # add model graph



import os
import shutil


def delete_directory(directory_path):
  if os.path.exists(directory_path):
    shutil.rmtree(directory_path)
    print(f"Deleted {directory_path}")
  else:
    print(f"Directory {directory_path} does not exist.")
  return


USELESS_PLUGINS = [
  'VIEW_SCENE',
  'REST_CUSTOM_EXEC'
]


def useless_dir(directory_name):
  for plugin in USELESS_PLUGINS:
    if plugin in directory_name:
      return True
  return False


def relevant_dir(directory_name):
  return dir_name.startswith('__') and 'archive_output' not in directory_name


def extract_ts_from_path(path):
  ts_str = path.split('_')[0]
  return ts_str


root_dir = r'C:\repos\edge-node\_local_cache\_output'
# res = 'lowres_take7_raise055-050'
res = 'lowres_640x1152_take6_065_050'
res_dest = 'highres_take2'
start_ts = '20240228120347000000'


if __name__ == '__main__':
  dir_list = []
  # Clearing the output folder
  for dir_name in os.listdir(root_dir):
    if os.path.isdir(os.path.join(root_dir, dir_name)):
      abs_dir = os.path.join(root_dir, dir_name)
      if useless_dir(dir_name):
        delete_directory(abs_dir)
      else:
        if relevant_dir(dir_name):
          dir_list.append(abs_dir)
        # end if good dir
      # end if remaining dir
    # end for dirs
  # end for os.walk

  if True:
    for dir_name in dir_list:
      os.makedirs(os.path.join(dir_name, res), exist_ok=True)
      for path in os.listdir(dir_name):
        if os.path.isfile(os.path.join(dir_name, path)):
          abs_path = os.path.join(dir_name, path)
          new_path = os.path.join(dir_name, res, path)
          shutil.move(abs_path, new_path)
        # end if file
      # end for path
    # end for dir_name

  if False:
    for dir_name in dir_list:
      curr_dir = os.path.join(dir_name, res)
      os.makedirs(os.path.join(dir_name, res_dest), exist_ok=True)

      for path in os.listdir(curr_dir):
        if os.path.isfile(os.path.join(curr_dir, path)):
          abs_path = os.path.join(curr_dir, path)
          ts = extract_ts_from_path(path)
          new_path = os.path.join(dir_name, res_dest, path)
          if ts > start_ts:
            shutil.move(abs_path, new_path)
          # endif move
        # end if file
      # end for path
    # end for dir_name




"""
TODO: OBSOLETE move/use from LIBS !
"""

from naeural_core.local_libraries.llm_api.openai.app import OpenAIApp

PERSONAS_FOLDER = './plugins/utils/openai/personas'
  
if __name__ == '__main__':
  DATA_TEST = False
  CODEGEN = True

  if DATA_TEST:
    query = 'Please generate in json format quizes where the answer is single word and completes the question. Propose another 4 wrong asnwers. Thematic is elementary school math. The questions should be amusing. Simplify the json format such as the "question" is "q", "correct_answer" is "a", "wrong_answers" is "wa" and the language "l":"en" or "l":"ro". Write the json puting each object in one line with the keys in the following order: "l", "q", "a", "wa" and end the line with a comma. Generate a total of 50 quizes for Romanian language'
    eng = OpenAIApp(persona="Worker", user='Andrei')
    res = eng.ask_direct(query)
    print(res)
    
  if CODEGEN:
    eng1 = OpenAIApp(persona='codegen', user='a1', persona_location=PERSONAS_FOLDER)
    r11 = eng1.ask("Write a simple hello world program in c++")
    eng2 = OpenAIApp(persona='codegen', user='a1', persona_location=PERSONAS_FOLDER)    
    r21 = eng2.ask("Write a simple hello world program in python")

    r12 = eng1.ask("Modify the program so that it prints the current date and time")
    r22 = eng2.ask("Modify the program so that it prints the system memory usage")
    
    print(r11)
    print("***********")
    print(r12)
    print("######################################################")
    print("######################################################")
    print(r21)
    print("***********")
    print(r22)



import numpy as np
from collections import deque
from naeural_core.business.templates.counting_template_mixin import _CountingTemplateMixin

from naeural_core.utils.basic_anomaly_model import BasicAnomalyModel


class _TimeBinsMixin(object):
  # TODO: change name to api calls (tbinapi_{})
  def __init__(self, **kwargs):
    # self.__time_bins = {} TODO: This doesn't work, don't know why
    self._time_bins = {}
    self._accepted_functions = {
      "mean": self.__custom_mean,
      "median": self.__custom_median,
      "count": self.__custom_count,
      "min": self.__custom_min,
      "max": self.__custom_max,
      "std": self.__custom_std,
      "sum": self.__custom_sum,
    }
    super(_TimeBinsMixin, self).__init__(**kwargs)
    return

  def __custom_mean(self, data):
    if len(data) == 0:
      return self.cfg_report_default_empty_value
    return np.mean(data)

  def __custom_median(self, data):
    if len(data) == 0:
      return self.cfg_report_default_empty_value
    return np.median(data)

  def __custom_count(self, data):
    return len(data)

  def __custom_min(self, data):
    if len(data) == 0:
      return self.cfg_report_default_empty_value
    return min(data)

  def __custom_max(self, data):
    if len(data) == 0:
      return self.cfg_report_default_empty_value
    return max(data)

  def __custom_std(self, data):
    if len(data) == 0:
      return self.cfg_report_default_empty_value
    return np.std(data)

  def __custom_sum(self, data):
    if len(data) == 0:
      return self.cfg_report_default_empty_value
    return sum(data)

  def timebins_create_bin(self, key="default", weekday_names=None, report_default_empty_value=None, per_day_of_week_timeslot=None, warmup_anomaly_models=None):
    if key not in self._time_bins:
      self._time_bins[key] = _TimeBins(
        self,
        weekday_names if weekday_names is not None else self.cfg_weekday_names,
        report_default_empty_value if report_default_empty_value is not None else self.cfg_report_default_empty_value,
        per_day_of_week_timeslot if per_day_of_week_timeslot is not None else self.cfg_per_day_of_week_timeslot,
        warmup_anomaly_models if warmup_anomaly_models is not None else self.cfg_warmup_anomaly_models
      )
    return

  def _save_bins(self):
    # self.persistence_serialization_save(self._deques)
    return

  def _maybe_load_bins(self):
    # try:
    #   deq = self.persistence_serialization_load()
    # except:
    #   return
    # if deq is None:
    #   return
    # self._get_bin()

    # def _merge(source, destination):
    #   """ Deep merge two dicts.

    #   Args:
    #       source (dict):
    #       destination (dict):

    #   Returns:
    #       dict: the resulting dict after merge
    #   """
    #   for key, value in source.items():
    #       if isinstance(value, dict):
    #           # get node or create one
    #           node = destination.setdefault(key, {})
    #           _merge(value, node)
    #       else:
    #           destination[key] = value

    #   return destination

    # self._deques = _merge(deq, self._deques)
    return

  def timebins_append(self, value, key='default'):
    assert key in self._time_bins, "Unknown time_bins key {}".format(key)
    self._time_bins[key]._append_bin(value)

  def timebins_get_bin(self, key='default'):
    assert key in self._time_bins, "Unknown time_bins key {}".format(key)
    return self._time_bins[key]._get_bin()

  def timebins_get_bin_mean(self, key='default'):
    assert key in self._time_bins, "Unknown time_bins key {}".format(key)
    return self._time_bins[key]._get_bin_mean()

  def timebins_is_anomaly(self, key='default'):
    assert key in self._time_bins, "Unknown time_bins key {}".format(key)
    return self._time_bins[key]._is_binned_anomaly()

  def timebins_get_bin_size(self, key='default'):
    assert key in self._time_bins, "Unknown time_bins key {}".format(key)
    return self._time_bins[key]._get_bin_size()

  def timebins_get_total_size(self, key='default'):
    assert key in self._time_bins, "Unknown time_bins key {}".format(key)
    return self._time_bins[key]._get_total_size()

  def timebins_get_bin_report(self, aggr_func="mean", key='default'):
    assert key in self._time_bins, "Unknown time_bins key {}".format(key)
    # TODO: base on the string, compute mean, median, count, min, max, std, sum
    assert aggr_func in self._accepted_functions
    return self._time_bins[key]._get_bin_report(aggr_func, self._accepted_functions[aggr_func])

  def timebins_get_bin_statistic(self, aggr_func="mean", key='default'):
    assert key in self._time_bins, "Unknown time_bins key {}".format(key)
    assert aggr_func in self._accepted_functions

    return self._time_bins[key]._get_bin_aggr_func(self._accepted_functions[aggr_func])

  def timebins_get_previous_bin_statistic(self, aggr_func="mean", key='default'):
    assert key in self._time_bins, "Unknown time_bins key {}".format(key)
    assert aggr_func in self._accepted_functions

    return self._time_bins[key]._get_previous_bin_aggr_func(self._accepted_functions[aggr_func])


class _TimeBins(object):
  # TODO: add to base?
  def __init__(self, owner, weekday_names, report_default_empty_value, per_day_of_week_timeslot, warmup_anomaly_models, **kwargs):
    self._deques = {}
    self._anomaly_detectors = {}
    self._owner = owner

    self.weekday_names = weekday_names
    self.report_default_empty_value = report_default_empty_value
    self.per_day_of_week_timeslot = per_day_of_week_timeslot
    self.warmup_anomaly_models = warmup_anomaly_models

    return

  def startup(self):
    super().startup()

  # @property
  # def cfg_weekday_names(self):
  #   return self._instance_config.get("WEEKDAY_NAMES", ['Monday', 'Tuesday', 'Wednesday', 'Thursday', 'Friday', 'Saturday', 'Sunday'])

  # @property
  # def cfg_report_default_empty_value(self):
  #   return self._instance_config.get("REPORT_DEFAULT_EMPTY_VALUE", -1)

  # @property
  # def cfg_per_day_of_week_timeslot(self):
  #   return self._instance_config.get("PER_DAY_OF_WEEK_TIMESLOT", True)

  # @property
  # def cfg_warmup_anomaly_models(self):
  #   return self._instance_config.get("WARMUP_ANOMALY_MODELS", 10)

  def _maybe_init_bin(self, weekday, interval):
    intervals = self._owner._calculate_daily_intervals()
    index = interval
    next_index = (index + 1) % len(intervals)
    i = intervals[index]
    next_i = intervals[next_index]

    if self.per_day_of_week_timeslot:
      if weekday not in self._deques or weekday not in self._anomaly_detectors:
        self._deques[weekday] = {}
        self._anomaly_detectors[weekday] = {}
      # endif

      if interval not in self._deques[weekday] or interval not in self._anomaly_detectors[weekday]:
        self._deques[weekday][interval] = _TimeBinDeque(
          maxlen=20000, interval=f"{self.weekday_names[weekday][:3]}_{i}-{next_i}")
        self._anomaly_detectors[weekday][interval] = BasicAnomalyModel()
      # endif
    else:
      if interval not in self._deques or interval not in self._anomaly_detectors:
        self._deques[interval] = _TimeBinDeque(
          maxlen=20000, interval=f"{i}-{next_i}")
        self._anomaly_detectors[interval] = BasicAnomalyModel()
      # endif
    # endif

  def _get_bin(self, interval=None, weekday=None):
    # TODO: should receive a now_time
    i, w = self._owner._check_daily_interval()
    interval = interval if interval is not None else i
    weekday = weekday if weekday is not None else w

    self._maybe_init_bin(
      interval=interval,
      weekday=weekday
    )
    if self.per_day_of_week_timeslot:
      return self._deques[weekday][interval]
    else:
      return self._deques[interval]
    # endif
    return

  def _get_bin_mean(self, interval=None, weekday=None):
    dq = self._get_bin(interval, weekday)
    if dq is not None:
      return np.mean(dq)
    return None

  def _init_all_bins(self):
    daily_intervals = self._owner._calculate_daily_intervals()
    for interval in range(len(daily_intervals)):
      if self.per_day_of_week_timeslot:
        for weekday in range(7):
          self._maybe_init_bin(
            interval=interval,
            weekday=weekday
          )
      else:
        _, weekday = self._owner._check_daily_interval()
        self._maybe_init_bin(
          interval=interval,
          weekday=weekday
        )

  def __format_report(self, report, aggr_func_name=None):
    formatted_report = {}

    if self.per_day_of_week_timeslot:
      for weekday, intervals in report.items():
        intervals_names = list(intervals.keys())
        weekday_name = self.weekday_names[weekday]
        formatted_report[weekday_name] = {}
        for i in range(len(intervals) - 1):
          value = intervals[intervals_names[i]]

          formatted_report[weekday_name][intervals_names[i]] = {
            'START': intervals_names[i],
            'END': intervals_names[i + 1],
            'VALUE': value,
            'AGGR_FUNC': aggr_func_name
          }
    else:
      formatted_report = {}
      intervals_names = list(report.keys())
      for i in range(len(report) - 1):
        value = report[intervals_names[i]]

        formatted_report[intervals_names[i]] = {
          'START': intervals_names[i],
          'END': intervals_names[i + 1],
          'VALUE': value,
          'AGGR_FUNC': aggr_func_name
        }
    return formatted_report

  def __format_result(self, value, i):
    intervals = self._owner._calculate_daily_intervals()
    index = i
    next_index = (index + 1) % len(intervals)
    i = intervals[index]
    next_i = intervals[next_index]

    start_interval = self._owner._interval_to_now_datetime(i)
    end_interval = self._owner._interval_to_now_datetime(next_i)

    start_interval = self._owner.log.time_to_str(start_interval.timestamp())
    end_interval = self._owner.log.time_to_str(end_interval.timestamp())

    return {'START': start_interval, 'END': end_interval, 'VALUE': value}

  def _get_bin_size(self, interval=None, weekday=None):
    dq = self._get_bin(interval, weekday)
    if dq is not None:
      return len(dq)
    return None

  def _get_total_size(self):
    total_size = 0
    for w in self._deques.values():
      for h in w.values():
        total_size += len(h)
      # endfor
    # endof
    return total_size

  def _get_previous_bin_aggr_func(self, aggr_func):
    i, w = self._owner._get_previous_interval()
    result = self._get_bin_aggr_func(aggr_func, interval=i, weekday=w)
    return self.__format_result(result, i)

  def _get_bin_aggr_func(self, aggr_func, interval=None, weekday=None):
    dq = self._get_bin(interval, weekday)
    if dq is not None:
      return aggr_func(dq)
    return None

  def _get_bin_report(self, aggr_func_name, aggr_func):
    report = {}
    daily_intervals = self._owner._calculate_daily_intervals()
    self._init_all_bins()
    if self.per_day_of_week_timeslot:
      for weekday in range(7):
        report[weekday] = {}
        for i, interval in enumerate(daily_intervals):
          report[weekday][interval] = self._get_bin_aggr_func(aggr_func, i, weekday)
        # end for intervals
      # end for weekdays
    else:
      for i, interval in enumerate(daily_intervals):
        report[interval] = self._get_bin_aggr_func(aggr_func, i)
      # end for intervals
    return self.__format_report(report, aggr_func_name)

  def _get_model(self):
    interval, weekday = self._owner._check_daily_interval()
    self._maybe_init_bin(interval=interval, weekday=weekday)
    if self.per_day_of_week_timeslot:
      return self._anomaly_detectors[weekday][interval]
    else:
      return self._anomaly_detectors[interval]
    # endif
    return

  def _append_bin(self, value):
    # TODO: should receive a now_time
    dq = self._get_bin()
    dq.append(value)

  def _is_binned_anomaly(self):
    dq = self._get_bin()
    if len(dq) > self.warmup_anomaly_models:
      model = self._get_model()
      model.fit(np.expand_dims(np.array(dq)[:-1], -1), prc=0.05)
      res = model.predict(np.expand_dims(np.array(dq)[-1], -1))
      return bool(res[0][0])
    return False


class _TimeBinDeque(deque):
  def __init__(self, *args, interval, **kwargs):
    self.interval = interval
    super(_TimeBinDeque, self).__init__(*args, **kwargs)

  def __str__(self):
    min_v = f"{np.min(self):.02f}" if len(self) > 0 else 'n'
    max_v = f"{np.max(self):.02f}" if len(self) > 0 else 'n'
    mean_v = f"{np.mean(self):.02f}" if len(self) > 0 else 'n'
    median_v = f"{np.median(self):.02f}" if len(self) > 0 else 'n'
    header = f"{self.interval} m/M/a/med={min_v}/{max_v}/{mean_v}/{median_v} |"
    return header + " ".join([f"{x}" if isinstance(x, int) else f"{x:.02f}" for x in list(self)[-10:]])


"""
no weekdays
[
  {
    'start': 'timp',
    'end': 'timp',
    'value': ___
  }
]

{
  'timp-timp': ____
}

with weekdays
{
  'monday':[
    {
      'start': 'timp',
      'end': 'timp',
      'value': ___
    }
  ]
}


"""


"""
report = {
  'start': {
    'start': __
    'end': ___
    'value': __
    'aggr_func': ___
  }
}

report = {
  'monday':{
    'start': {
      'start': __
      'end': ___
      'value': __
      'aggr_func': ___
    }
  }
}



"""



import paho.mqtt.subscribe as subscribe

DEVICE_ID = '12345'

_CONFIG_RECEIVER = {
  'MQTT_HOST': 'HOSTNAME',
  'MQTT_PORT': 1883,
  'MQTT_USER': 'USERNAME',
  'MQTT_PASS': 'PASSWORD',
  'MQTT_PATH': 'PATH',
  'MQTT_TOPIC': 'config1',
  'MQTT_QOS': 1
  }

_CONFIG = _CONFIG_RECEIVER

if __name__ == '__main__':  
  topic = '{}/{}'.format(_CONFIG['MQTT_PATH'], _CONFIG['MQTT_TOPIC'])
  
  while True:
    m = subscribe.simple(
      topics=topic, 
      qos=1, 
      hostname=_CONFIG['MQTT_HOST'],
      port=_CONFIG['MQTT_PORT'], 
      auth={'username': _CONFIG['MQTT_USER'], 'password': _CONFIG['MQTT_PASS']}
      )
    print(m.payload.decode('utf-8'))


import torch as th
import torchvision as tv



def th_resize_with_pad(img, h, w, device=None, float_convert_first=False):
  def _th_resize(th_img):
    if isinstance(th_img, np.ndarray):
      th_img = th.tensor(th_img, device=device)
    if th_img.shape[-1] == 3:
      if len(th_img.shape) == 3:
        th_img = th_img.permute((2,0,1))
      elif len(th_img.shape) == 4:
        th_img = th_img.permute((0, 3, 1, 2))
    if float_convert_first and th_img.dtype == th.uint8:
      th_img = th_img / 255
        
    fill_value = 144 if th_img.dtype == th.uint8 else 144/255
    new_shape = h, w
    shape = th_img.shape[-2:]  # current shape [height, width]
    r = min(new_shape[0] / shape[0], new_shape[1] / shape[1])
    new_unpad = int(round(shape[0] * r)), int(round(shape[1] * r))
    dw, dh = new_shape[1] - new_unpad[1], new_shape[0] - new_unpad[0]  # wh padding
    dw /= 2  # divide padding into 2 sides
    dh /= 2
    if shape != new_unpad:
      th_resized = tv.transforms.Resize(new_unpad)(th_img)
    else:
      th_resized = th_img
    top, bottom = int(round(dh - 0.1)), int(round(dh + 0.1))
    left, right = int(round(dw - 0.1)), int(round(dw + 0.1))
    th_resized_padded = tv.transforms.Pad((left, top, right, bottom), fill=fill_value)(th_resized)
    
    if th_img.dtype == th.uint8:
      th_img = th_img / 255
    return th_resized_padded
  
  if isinstance(img, list):
    images = []
    for im in img:
      images.append(_th_resize(im).unsqueeze(0))
    th_out = th.cat(images)
  else:
    th_out = _th_resize(img)
  return th_out
  

if __name__ == '__main__':
  import matplotlib.pyplot as plt
  import numpy as np
  from PIL import Image
  
  orig_fns = [
    'xperimental/_images/H1080_W1920/faces5.jpg',
    'xperimental/_images/H1080_W1920/faces9.jpg',
    'xperimental/_images/H1080_W1920/faces21.jpg',
    'xperimental/_images/H1520_W2688/bmw_man1.png',
    'xperimental/_images/H2048_W3072/bmw_man1.png',
    ]
  
  all_imgs = [np.asarray(Image.open(x)) for x in orig_fns]
  
  def plot(orig_img, imgs, with_orig=True, row_title=None, **imshow_kwargs):
    if not isinstance(imgs[0], list):
        # Make a 2d grid even if there's just 1 row
        imgs = [imgs]

    num_rows = len(imgs)
    num_cols = len(imgs[0]) + with_orig
    fig, axs = plt.subplots(nrows=num_rows, ncols=num_cols, squeeze=False, figsize=(13,8))
    for row_idx, row in enumerate(imgs):
        row = [orig_img] + row if with_orig else row
        for col_idx, img in enumerate(row):
            ax = axs[row_idx, col_idx]
            img = np.transpose(img.squeeze(), (1,2,0))
            ax.imshow(img, **imshow_kwargs)
            ax.set(xticklabels=[], yticklabels=[], xticks=[], yticks=[])
            ax.set(title='{}x{}'.format(*img.shape[:2]))

    if with_orig:
        axs[0, 0].set(title='Original image {}x{}'.format(*orig_img.shape[-2:]))
        axs[0, 0].title.set_size(8)
    if row_title is not None:
        for row_idx in range(num_rows):
            axs[row_idx, 0].set(ylabel=row_title[row_idx])

    plt.tight_layout()  

  dev = th.device('cuda')

  np_imgs = all_imgs[:3]  
  th_imgs = th.tensor(np_imgs, device=dev)
  
  inputs = np_imgs
  th_resized_padded = th_resize_with_pad(inputs,h=1200,w=1920,float_convert_first=False, device=dev)

  # TODO: 
  #    - test resize -> float/255 vs float/255 -> resize (timeit)
  # 

  for i in range(3):
    plot(
      orig_img=all_imgs[i].transpose((2,0,1)), 
      imgs=[
        th_resized_padded[i].cpu().numpy(),
        ])
  



from naeural_core.data.base import DataCaptureThread

_CONFIG = {
  **DataCaptureThread.CONFIG,

  'CAP_RESOLUTION': 10,
  'LOG_CONNECTION_ERROR_PERIOD': 30,  # seconds

  'ISAPI_CONFIG_METADATA': {
    'MESSAGE_DELIMITER': '--boundary',
    'RESPONSE_AS_JSON': False,
    'XML_SCHEMA': None,  # this should be required only if we read xml's
    'ISAPI_ROUTE': None,
    'CAMERA_IP': None,
    'USERNAME': None,
    'PASSWORD': None,
  },

  'VALIDATION_RULES': {
    **DataCaptureThread.CONFIG['VALIDATION_RULES'],
  },
}


class IsapiStreamDataCapture(DataCaptureThread):
  CONFIG = _CONFIG

  def __init__(self, **kwargs):
    super(IsapiStreamDataCapture, self).__init__(**kwargs)
    self._iterator = None
    self.buffer = b''

    self.__last_connection_status_error_time = 0
    return

  def startup(self):
    super().startup()

    self._camera_ip = self.cfg_isapi_config_metadata.get('CAMERA_IP', None)
    self._username = self.cfg_isapi_config_metadata.get('USERNAME', None)
    self._password = self.cfg_isapi_config_metadata.get('PASSWORD', None)
    return

  @property
  def message_delimiter(self):
    delimiter = self.cfg_isapi_config_metadata.get('MESSAGE_DELIMITER', '--boundary')
    return bytes(delimiter, 'utf-8')

  @property
  def isapi_route(self):
    return self.cfg_isapi_config_metadata.get('ISAPI_ROUTE', None)

  @property
  def xml_schema(self):
    return self.cfg_isapi_config_metadata.get('XML_SCHEMA', None)

  @property
  def response_as_json(self):
    return self.cfg_isapi_config_metadata.get('RESPONSE_AS_JSON', False)

  def _init(self):
    self._maybe_reconnect()
    return

  def _maybe_reconnect(self):
    if self.has_connection:
      return

    url = f"http://{self._camera_ip}{self.isapi_route}"
    if self.response_as_json:
      url += '?format=json'

    if self._iterator is None:
      try:
        response = self.requests.get(
          url,
          auth=self.requests.auth.HTTPDigestAuth(self._username, self._password),
          stream=True
        )
        if response is None:
          self.P("Connection fail: remote did not respond", color='r')
          return

        if response.status_code != 200:
          if self.time() - self.__last_connection_status_error_time > self.cfg_log_connection_error_period:
            self.P(f"Connection fail: the status code is {response.status_code}", color='r')
            self.__last_connection_status_error_time = self.time()
          return
        self.__last_connection_status_error_time = 0

        # we use an iterator to read the response byte by byte
        # we no longer use response.iter_lines() because it is not reliable
        self._iterator = response.iter_content(chunk_size=16 * 1024)
        self.buffer = b''

        # this first acquire is to get the first message delimiter
        if self.acquire_full_http_response(parse_response=False) is not None:
          self.has_connection = True
          self.P("Connected to remote")
      except:
        self.P("Connection fail: timeout", color='r')
    return

  def _extract_xml_from_content(self, content, **kwargs):
    """
    Extracts the xml from the content.

    Parameters
    ----------
    content : bytes
        The content to extract the xml from.

    Returns
    -------
    xml : ElementTree
        The xml.
    """

    # we know that the xml starts with <EventNotificationAlert>
    start = 0

    # find the end of the xml
    end = content.find(b'</EventNotificationAlert>')
    if end == -1:
      self.P("Connection fail: end of xml not found", color='r')
      return

    # extract the xml
    xml_content = content[start:end + len('</EventNotificationAlert>\r\n')]

    # replace the xml schema if it is provided
    xml_content = xml_content.replace(bytes(f' xmlns="{self.xml_schema}"', "utf-8"), b'')

    xml = self.ElementTree.fromstring(xml_content)

    return xml

  def _extract_image_from_content(self, content, read_plate=False, **kwargs):
    """
    Extracts the image from the content.

    Parameters
    ----------
    content : bytes
        The content to extract the image from.

    Returns
    -------
    image_plate_tlbr : Tuple[PIL.Image, Tuple[float, float, float, float]|None]
        The image and the plate tlbr.
    """

    # # find the start of the image
    # start == 0 always

    # # find the end of the image
    # end == len(content) - 2 always

    # extract the image
    self.start_timer('anpr_img_extract')
    image_content = content[:-2]

    image = self.BytesIO(image_content)

    pil_image = self.PIL.Image.open(image)
    self.end_timer('anpr_img_extract')

    return pil_image

  def _extract_json_from_content(self, content, **kwargs):
    """
    Extracts JSON from the content.

    Parameters
    ----------
    content : bytes
        The content to extract JSON from.

    Returns
    -------
    object : dict
        The extracted JSON object.
    """

    # TODO(S): test this
    self.P("WARNING: JSON parsing is not tested", color='r')

    # we know that the json starts at the first byte
    start = 0

    # find the end of the json
    end = content.find(b'}')

    # extract the json
    json_content = content[start:end + 1]

    dct_json = self.json.loads(json_content)

    return dct_json

  def extract_object_from_response(self, dct_response, **kwargs):
    """
    Extracts the object from the response.

    Parameters
    ----------
    dct_response : dict
        The response to extract the object from.

    Returns
    -------
    object : ElementTree or Tuple[PIL.Image, Tuple[float, float, float, float]|None]
    """

    content_type = dct_response['content_type']
    content = dct_response['content']

    obj = None

    if b'xml' in content_type:
      self.start_timer('anpr_xml_parse')
      obj = self._extract_xml_from_content(content, **kwargs)
      self.end_timer('anpr_xml_parse')

    elif b'json' in content_type:
      self.start_timer('anpr_json_parse')
      obj = self._extract_json_from_content(content, **kwargs)
      self.end_timer('anpr_json_parse')

    elif b'image/jpeg' in content_type:
      self.start_timer('anpr_img_parse')
      obj = self._extract_image_from_content(content, **kwargs)
      self.end_timer('anpr_img_parse')
    return obj

  def acquire_full_http_response(self, parse_response=True):
    """
    Acquires the full http response.

    Returns
    -------
    dct_response : dict or None
        The parsed response, containing the content type and the content.
    """

    self.start_timer('anpr_http')

    dct_response = None

    index = self.buffer.find(self.message_delimiter)
    if index != -1:
      response = self.buffer[:index]
      dct_response = self._maybe_parse_http_response(response, parse_response=parse_response)
      self.buffer = self.buffer[index + len(self.message_delimiter):]
      self.end_timer('anpr_http')
      return dct_response

    content_len = None
    current_content_len = 0
    lst_byte_chunks = []

    try:
      while True:
        if self._iterator is None:
          return dct_response
        self.start_timer('anpr_http_next')
        byte_chunk = next(self._iterator)
        self.end_timer('anpr_http_next')

        # read bytes until hitting the message delimiter
        # we must make sure we do not skip delimiters split between chunks
        # and so we can update the buffer with the new bytes
        # (this method is not very efficient but it is reliable)

        if content_len is None:
          # break early if we have already found the message delimiter
          index = self.buffer.find(self.message_delimiter)
          if index != -1:
            response = self.buffer[:index]
            dct_response = self._maybe_parse_http_response(response, parse_response=parse_response)
            self.buffer = self.buffer[index + len(self.message_delimiter):] + byte_chunk
            break

          # find the end of the header
          # the header is small so searching from the start is not a big problem
          self.start_timer('anpr_http_header')
          header_end_index = self.buffer.find(b'\r\n\r\n')
          self.end_timer('anpr_http_header')

          if header_end_index == -1:
            # we haven't finished reading the header yet
            self.buffer += byte_chunk
            continue

          # first time computing content_len
          for header in self.buffer.split(b'\r\n'):
            # we know we have the entire header
            if header.startswith(b'Content-Length'):
              content_len = int(header.split(b':')[1].strip())
              # we add the length of the separator to the content length because
              # we determine the end of the content with it
              content_len += len(self.message_delimiter) + len('\r\n\r\n')
            if header.startswith(b'Content-Type'):
              content_type = header.split(b':')[1].strip()
          # endfor

          current_content_len = len(self.buffer[header_end_index + len('\r\n\r\n'):])

        lst_byte_chunks.append(byte_chunk)
        current_content_len += len(byte_chunk)

        if current_content_len >= content_len:
          self.start_timer('anpr_http_join')
          self.buffer += b''.join(lst_byte_chunks)
          self.end_timer('anpr_http_join')
          lst_byte_chunks = []

          index = self.buffer.find(self.message_delimiter, header_end_index + content_len - len(self.message_delimiter))
          # because we added the length of the separator to the content length we should always find the separator
          if index != -1:
            response = self.buffer[:index]
            dct_response = {
              'content_type': content_type,
              'content': response[header_end_index + len('\r\n\r\n'):],
            }
            self.buffer = self.buffer[index + len(self.message_delimiter):]
            break
      # endfor

    except Exception as e:
      self._iterator = None
      self.has_connection = False
      self.P(f"Some error occured: {e}")
    self.end_timer('anpr_http')
    return dct_response

  def _maybe_parse_http_response(self, response, parse_response=True):
    """
    Parses the http response.

    Parameters
    ----------
    response : bytes
        The response to parse.

    Returns
    -------
    dct_response : dict
        The parsed response, containing the content type and the content.
    """

    if not parse_response:
      return response

    content_type = None
    for header in response.split(b'\r\n'):
      if header.startswith(b'Content-Type'):
        content_type = header.split(b':')[1].strip()
        break

    if content_type is None:
      self.P("Connection fail: content type is None", color='r')
      return

    # find the content
    content = None
    start = response.find(b'\r\n\r\n')
    content = response[start + len('\r\n\r\n'):]

    if content is None:
      self.P("Connection fail: content is None", color='r')
      return

    dct_response = {
      'content_type': content_type,
      'content': content,
    }

    return dct_response

  def _run_data_aquisition_step(self):
    self.start_timer('isapi_http')
    dct_response = self.acquire_full_http_response()
    self.end_timer('isapi_http')

    if dct_response is None:
      self.P("Connection fail: remote did not respond", color='r')
      return

    self.start_timer('isapi_get_object')
    obj = self.extract_object_from_response(dct_response)
    self.end_timer('isapi_get_object')

    if b'image' in dct_response['content_type']:
      image = self.np.array(obj)[::-1]
      self._add_img_input(image)

    else:
      self._add_struct_data_input({
        'content_type': dct_response['content_type'],
        'content': obj,
      })

    return

  def _release(self):
    self._iterator = None
    return



"""


FinalClass.__mro__ = (<class '__main__.FinalClass'>, <class '__main__.BaseClass'>, <class '__main__._Mixin1'>, <class '__main__._Mixin2'>, <class '__main__._Mixin3'>, <class 'object'>)
FinalClass.1
BaseClass.1 
_Mixin1.1   
_Mixin2.1   
_Mixin3.1   
_Mixin3.2   
_Mixin2.2   
_Mixin1.2   
BaseClass.2 
FinalClass.2
  
"""
class _Mixin1:
  def __init__(self):      
    print("_Mixin1.1".format(self.__class__.__name__))
    super(_Mixin1, self).__init__()
    print("_Mixin1.2".format(self.__class__.__name__))
    return
  
class _Mixin2:
  def __init__(self):      
    print("_Mixin2.1".format(self.__class__.__name__))
    super(_Mixin2, self).__init__()
    print("_Mixin2.2".format(self.__class__.__name__))
    return
  
  
class _Mixin3:
  def __init__(self):      
    print("_Mixin3.1".format(self.__class__.__name__))
    super(_Mixin3, self).__init__()
    print("_Mixin3.2".format(self.__class__.__name__))
    return
  

class BaseClass:
  def __init__(self):
    print("BaseClass.1".format(self.__class__.__name__))
    super(BaseClass, self).__init__()
    print("BaseClass.2".format(self.__class__.__name__))
    return
  
class FinalClass(BaseClass, _Mixin1, _Mixin2, _Mixin3):
  def __init__(self):
    print("FinalClass.1".format(self.__class__.__name__))
    super(FinalClass, self).__init__()
    print("FinalClass.2".format(self.__class__.__name__))
    return    
  
  
if __name__ == "__main__":
  print("FinalClass.__mro__ = {}".format(FinalClass.__mro__))
  eng = FinalClass()


# YOLOv5 🚀 by Ultralytics, GPL-3.0 license
"""
PyTorch Hub models https://pytorch.org/hub/ultralytics_yolov5/

Usage:
    import torch
    model = torch.hub.load('ultralytics/yolov5', 'yolov5s')
"""

import torch


def _create(name, pretrained=True, channels=3, classes=80, autoshape=True, verbose=True, device=None):
    """Creates a specified YOLOv5 model

    Arguments:
        name (str): name of model, i.e. 'yolov5s'
        pretrained (bool): load pretrained weights into the model
        channels (int): number of input channels
        classes (int): number of model classes
        autoshape (bool): apply YOLOv5 .autoshape() wrapper to model
        verbose (bool): print all information to screen
        device (str, torch.device, None): device to use for model parameters

    Returns:
        YOLOv5 pytorch model
    """
    from pathlib import Path

    from models.yolo import Model
    from models.experimental import attempt_load
    from utils.general import check_requirements, set_logging
    from utils.downloads import attempt_download
    from utils.torch_utils import select_device

    file = Path(__file__).resolve()
    check_requirements(exclude=('tensorboard', 'thop', 'opencv-python'))
    set_logging(verbose=verbose)

    save_dir = Path('') if str(name).endswith('.pt') else file.parent
    path = (save_dir / name).with_suffix('.pt')  # checkpoint path
    try:
        device = select_device(('0' if torch.cuda.is_available() else 'cpu') if device is None else device)

        if pretrained and channels == 3 and classes == 80:
            model = attempt_load(path, map_location=device)  # download/load FP32 model
        else:
            cfg = list((Path(__file__).parent / 'models').rglob(f'{name}.yaml'))[0]  # model.yaml path
            model = Model(cfg, channels, classes)  # create model
            if pretrained:
                ckpt = torch.load(attempt_download(path), map_location=device)  # load
                msd = model.state_dict()  # model state_dict
                csd = ckpt['model'].float().state_dict()  # checkpoint state_dict as FP32
                csd = {k: v for k, v in csd.items() if msd[k].shape == v.shape}  # filter
                model.load_state_dict(csd, strict=False)  # load
                if len(ckpt['model'].names) == classes:
                    model.names = ckpt['model'].names  # set class names attribute
        if autoshape:
            model = model.autoshape()  # for file/URI/PIL/cv2/np inputs and NMS
        return model.to(device)

    except Exception as e:
        help_url = 'https://github.com/ultralytics/yolov5/issues/36'
        s = 'Cache may be out of date, try `force_reload=True`. See %s for help.' % help_url
        raise Exception(s) from e


def custom(path='path/to/model.pt', autoshape=True, verbose=True, device=None):
    # YOLOv5 custom or local model
    return _create(path, autoshape=autoshape, verbose=verbose, device=device)


def yolov5s(pretrained=True, channels=3, classes=80, autoshape=True, verbose=True, device=None):
    # YOLOv5-small model https://github.com/ultralytics/yolov5
    return _create('yolov5s', pretrained, channels, classes, autoshape, verbose, device)


def yolov5m(pretrained=True, channels=3, classes=80, autoshape=True, verbose=True, device=None):
    # YOLOv5-medium model https://github.com/ultralytics/yolov5
    return _create('yolov5m', pretrained, channels, classes, autoshape, verbose, device)


def yolov5l(pretrained=True, channels=3, classes=80, autoshape=True, verbose=True, device=None):
    # YOLOv5-large model https://github.com/ultralytics/yolov5
    return _create('yolov5l', pretrained, channels, classes, autoshape, verbose, device)


def yolov5x(pretrained=True, channels=3, classes=80, autoshape=True, verbose=True, device=None):
    # YOLOv5-xlarge model https://github.com/ultralytics/yolov5
    return _create('yolov5x', pretrained, channels, classes, autoshape, verbose, device)


def yolov5s6(pretrained=True, channels=3, classes=80, autoshape=True, verbose=True, device=None):
    # YOLOv5-small-P6 model https://github.com/ultralytics/yolov5
    return _create('yolov5s6', pretrained, channels, classes, autoshape, verbose, device)


def yolov5m6(pretrained=True, channels=3, classes=80, autoshape=True, verbose=True, device=None):
    # YOLOv5-medium-P6 model https://github.com/ultralytics/yolov5
    return _create('yolov5m6', pretrained, channels, classes, autoshape, verbose, device)


def yolov5l6(pretrained=True, channels=3, classes=80, autoshape=True, verbose=True, device=None):
    # YOLOv5-large-P6 model https://github.com/ultralytics/yolov5
    return _create('yolov5l6', pretrained, channels, classes, autoshape, verbose, device)


def yolov5x6(pretrained=True, channels=3, classes=80, autoshape=True, verbose=True, device=None):
    # YOLOv5-xlarge-P6 model https://github.com/ultralytics/yolov5
    return _create('yolov5x6', pretrained, channels, classes, autoshape, verbose, device)


if __name__ == '__main__':
    model = _create(name='yolov5s', pretrained=True, channels=3, classes=80, autoshape=True, verbose=True)  # pretrained
    # model = custom(path='path/to/model.pt')  # custom

    # Verify inference
    import cv2
    import numpy as np
    from PIL import Image
    from pathlib import Path

    imgs = ['data/images/zidane.jpg',  # filename
            Path('data/images/zidane.jpg'),  # Path
            'https://ultralytics.com/images/zidane.jpg',  # URI
            cv2.imread('data/images/bus.jpg')[:, :, ::-1],  # OpenCV
            Image.open('data/images/bus.jpg'),  # PIL
            np.zeros((320, 640, 3))]  # numpy

    results = model(imgs)  # batched inference
    results.print()
    results.save()



"""
  This mixin can be used in order to automatically create custom datasets
  based on one or more business plugins.
  The plugins function by having inferences as an input and payload or payloads as output.
  To make the dataset this mixin will retrieve data both from the inferences and the payload.

  For reference, we will take the PerimeterIntrusion01 plugin. It is designed to analyze a
  CCTV video stream of a location and report any human that passes through the monitored area
  using a general detector.
  As the input the plugin will have the inferences packed in the following format:
  [
    {
      'PROB_PRC': 0.7, # the confidence of the current detection
      'TLBR_POS': [100, 230, 432, 316], # the localization of the current detection
      'TYPE': 'person', # class name of the current detection
      ...
    },
    ...
  ]
  The plugin will also produce a payload in the following format:
  {
    'IMG': numpy image, # witness image produced by the plugin
    'IMG_ORIG': numpy image, # original image on which the inference was computed
    ...
  }

  In the following we will analyze 3 levels of use for this Mixin.
  1) Beginner:
  For example, let's say we have a plugin, WeaponAssailant02, deployed on a CCTV camera that
  uses firstly a general detector in order to detect all the persons in
  the current frame and then a second model that tells us whether a person has a
  weapon in possession or not.
  This mixin can be used to crop the persons from all the images and save the crops
  along with each cropped image's prediction by the current working model.

  {
    ...
    "PLUGINS": [
        ...
        {
            "INSTANCES": [
                {
                    "INSTANCE_ID": "example dataset",
                    "DATASET_BUILDER": {
                      "DATASET_NAME": "weapon_1",
                      "LABEL_FILE_TEMPLATE": "naeural_core/xperimental/dataset_builder/weapon_label_template.txt"
                      "INFERENCE_MAPPING": {
                        "IS_ARMED": "TYPE"
                      }
                    }
                }
            ],
            "SIGNATURE": "WEAPON_ASSAILANT_02"
        }
        ...
    ],
    ...
  }

  In the above provided template for the label files we have:

  <annotation>
  <size>
    <width>##IMG_WIDTH##</width>
    <height>##IMG_HEIGHT##</height>
  </size>
  <object>
    <label>##IS_ARMED##</label>
  </object>
  </annotation>

"""

# TODO: in future versions in case we are referring to a list in a tag name we should
#  be able to specify the i-th element of that list (example: ##TLBR_POS:0## to specify the top coordinate)
# TODO: maybe not run ds builder at every main loop and add RESOLUTION parameter
# TODO: use witness_period
DEFAULT_DATASET_BUILDER_PARAMS = {
  "LABEL_FILE_TEMPLATE": None,
  "LABEL_TEMPLATE_TYPE": "FILE",  # TODO url
  "LABEL_EXTENSION": 'csv',
  "X_EXTENSION": "jpg",
  # TODO: maybe make x accept a template too (for non-images)
  "IMAGE_CROP": ["TLBR_POS"],  # parameters to use for cropping
  "IMAGE_SAVE": ["IMG_ORIG"],  # images from where to crop? | maybe dictionary in order to
  # specify different cropping coordinates for multiple images

  # in this we will specify additional keys that should be stored
  # from either the inferences or the payload
  "ADDITIONAL_KEYS": [],

  # in this we will specify keys that if missing the datapoint won't be saved
  "MANDATORY_KEYS": [],

  "INFERENCE_MAPPING": {},  # dictionary used if we want to use aliases for certain values from
  # either the model serving or from the payload (maybe we want to reduce the size of the template file)

  "DATASET_NAME": None,  # name of dataset
  "GENERIC_PATH": '',
  "EXPAND_VALUE": 0,  # in case of cropping from any images we can use this parameter to enlarge the crop
  # for example we may want to crop all the persons in a frame, but we want to save more of the scene than just
  # the person's detection box
  # If this parameter is > 1 it will be considered as the absolute value, in pixels, of the expansion
  # If this is <=1 it will be considered as a relative value and we will expand the cropped image by
  # ceil(expand_value * w) on the sides and ceil(expand_value * h) on the top and bottom, where
  # h, w are the height and width of the detection
  "OBJECT_MAX_SAVES": 0,  # this will determine how many appearances of a certain object will be saved
  # if this is 0 all the appearances will be saved
  # if the ai_engine used for the dataset creation does not use the tracker this will not be effective
  "TOTAL_MAX_SAVES": 0,
  "WITNESS_PERIOD": -1,  # if this x >= 0 this mixin will produce at most a witness every x seconds
  "ZIP_PERIOD": 0,  # after how many saved datapoints to archive the saved files in a zip,
  "STATS_UPDATE_PERIOD": 20,  # after how many saved datapoints to update the cached stats of the data saved so far
  # this will be used in order to continue the dataset creation in case of restarts without ignoring the already
  # saved data
  "SAVE_CAP_RESOLUTION": 0,  # from maximum how many frames per second to save datapoints
  "BACKGROUND_PERIOD_SAVE": 0,  # if this is > 0 we will save a background image every x processed frames
}


class _DatasetBuilderMixin(object):
  def __init__(self):
    self.__first_timestamp = None
    self.__dir_cnt = 0
    self.__saved_data_counter = 0
    self.__object_saves = {}
    self._ds_params_inferences = []
    self._ds_params_frame = {}
    self._template_fragments = None
    self._path_elements = None
    self._label_keys = set()
    self._mandatory_keys = set()
    self._last_save_ts = None
    self._is_multilabel = False
    self._background_saves = 0
    super(_DatasetBuilderMixin, self).__init__()
    return

  """VALID SECTION"""
  if True:
    def _valid_label_file_type(self, file_type):
      valid_types = self.ds_consts.VALID_FILE_TYPES
      if file_type not in valid_types:
        self.add_error(
          f"\"LABEL_TEMPLATE_TYPE\" parameter from "
          f"\"DATASET_BUILDER\" should be one of the "
          f"following: {valid_types}, but instead is {file_type}"
        )
      # TODO: validate correct format of the template file
      return

    def _valid_ds_name(self, ds_name):
      if not isinstance(ds_name, str):
        self.add_error(
          f"\"DATASET_NAME\" parameter from \"DATASET_BUILDER\" should be a string, but instead is {type(ds_name)}"
        )
      return

    def _valid_period(self, period_key, period_value, strict=False):
      if not isinstance(period_value, (int, float)):
        self.add_error(
          f"\"{period_key}\" parameter from \"DATASET_BUILDER\" should be an integer or a float,"
          f" but instead is {type(period_value)}"
        )
      else:
        value_check = period_value <= 0 if strict else period_value < 0
        requirement = 'strictly positive' if strict else 'positive'
        if value_check:
          self.add_error(
            f"\"{period_key}\" parameter from \"DATASET_BUILDER\" should be a {requirement} number,"
            f" but instead is {period_value}"
          )
        # endif value_check
      return

    def get_dataset_name(self):
      return self.get_dataset_builder_params.get(self.ds_consts.DATASET_NAME)

    def validate_dataset_builder(self):
      self.P('Validating DATASET_BUILDER')
      # this is reset in order to reparse it in case the template file was changed
      self._template_fragments = None
      self._path_elements = None
      ds_builder = self.cfg_dataset_builder
      if ds_builder is None:
        return

      current_params = self.get_dataset_builder_params
      self._valid_label_file_type(current_params.get(self.ds_consts.LABEL_TEMPLATE_TYPE, self.ds_consts.FILE))
      self._valid_ds_name(self.get_dataset_name())
      self._valid_period(period_key='ZIP_PERIOD', period_value=self.get_zip_period)
      self._valid_period(period_key='STATS_UPDATE_PERIOD', period_value=self.get_stats_update_period, strict=True)
      if self.__first_timestamp is None:
        self.__first_timestamp = self.log.now_str(nice_print=False, short=True)
      return
  """END VALID SECTION"""

  """GETTER SECTION"""
  if True:
    @property
    def get_plugin_default_dataset_builder_params(self):
      """
      Method that will be used for the plugins where the dataset builder mixin will be enabled by default
      in order to facilitate the configuration of this mixin without ignoring the default ds builder params
      provided in the plugin default config.
      Returns
      -------

      """
      return {}

    @property
    def get_dataset_builder_params(self):
      ds_builder_params = self.cfg_dataset_builder
      if ds_builder_params is None:
        ds_builder_params = {}
      return {
        **DEFAULT_DATASET_BUILDER_PARAMS,
        **self.get_plugin_default_dataset_builder_params,
        **ds_builder_params
      }

    @property
    def get_label_template_type(self):
      return self.get_dataset_builder_params.get(self.ds_consts.LABEL_TEMPLATE_TYPE)

    @property
    def get_label_file_template(self):
      return self.get_dataset_builder_params.get(self.ds_consts.LABEL_FILE_TEMPLATE)

    @property
    def get_inference_mapping(self):
      return self.get_dataset_builder_params.get(self.ds_consts.INFERENCE_MAPPING, {})

    @property
    def get_object_max_saves(self):
      # string instead of constant in the event we change its name
      return self.get_dataset_builder_params.get("OBJECT_MAX_SAVES", 0)

    @property
    def get_image_save(self):
      # string instead of constant in the event we change its name
      return self.get_dataset_builder_params.get("IMAGE_SAVE", [])

    @property
    def get_image_crop(self):
      # string instead of constant in the event we change its name
      return self.get_dataset_builder_params.get("IMAGE_CROP", [])

    @property
    def get_additional_keys(self):
      # string instead of constant in the event we change its name
      return self.get_dataset_builder_params.get("ADDITIONAL_KEYS", [])

    @property
    def get_mandatory_keys(self):
      return self.get_dataset_builder_params.get("MANDATORY_KEYS", [])

    @property
    def get_generic_path(self):
      return self.get_dataset_builder_params.get("GENERIC_PATH", '')

    @property
    def get_expand_value(self):
      return self.get_dataset_builder_params.get("EXPAND_VALUE", 0)

    @property
    def get_total_max_saves(self):
      # string instead of constant in the event we change its name
      return self.get_dataset_builder_params.get("TOTAL_MAX_SAVES", 0)

    @property
    def get_zip_period(self):
      return self.get_dataset_builder_params.get("ZIP_PERIOD", 0)

    @property
    def get_stats_update_period(self):
      return self.get_dataset_builder_params.get("STATS_UPDATE_PERIOD", 20)

    @property
    def get_background_period_save(self):
      return self.get_dataset_builder_params.get("BACKGROUND_PERIOD_SAVE", 0)

    def get_label_template_lines(self):
      if self.get_label_template_type == self.ds_consts.FILE:
        if self.get_label_file_template is None:
          return []
        with open(self.get_label_file_template, 'r') as temp_in:
          return temp_in.readlines()
      # endif self.cfg_label_template_type == "FILE"

      return []

    def get_x_file_params(self):
      """
      This method will compute a list of tuples of 2 elements that describe what
      keys to use in order to compute and save data for the X files in our dataset.
      The above-mentioned keys will be found in the data provided either by
      the plugin or the model_serving(we will refer to that as dict).
      Can also be further extended by the plugin developer in order to customise
      the extracted params.
      Returns
      -------
        res - list of format
        [
          (SOURCE_KEY1, PROP_KEY1),
          ..
        ]
        where dict[PROP_KEYi] will be used to extract data from dict[SOURCE_KEYi]
      """
      res = []
      sources = self.get_image_save
      props = self.get_image_crop
      for source in sources:
        for prop in props:
          res.append((source, prop))
        # endfor props
      # endfor sources
      return res

    def get_save_cap_resolution(self):
      return self.get_dataset_builder_params.get('SAVE_CAP_RESOLUTION', 0)
  """END GETTER SECTION"""

  def maybe_init_ds_builder_saved_stats(self):
    if self.cfg_dataset_builder is None:
      return
    saved_stats = self.get_saved_stats()
    if saved_stats is not None:
      self.P('Save stats backup found. Loading saved stats for dataset builder mixin')
      self.__dir_cnt = saved_stats.get('dir_cnt', 0)
      self.__saved_data_counter = saved_stats.get('saved_data_counter', 0)
      self.__object_saves = saved_stats.get('object_saves', {})
      self.P(f'Saved stats loaded! Currently on chunk {self.__dir_cnt} after {self.__saved_data_counter} total saves.')
    # endif saved stats found
    return

  def __enough_saves(self):
    return self.get_total_max_saves > 0 and self.__saved_data_counter >= self.get_total_max_saves

  def _maybe_init_template_fragments(self):
    """
    Here we will parse the text from the template label file.
    The text in our template can either be normal text or the
    name of a tag(which will be enclosed in ##).
    Because of this if we will split one fragment of the above-mentioned file
    by '##' we will obtain an array of words indexed from 0 where on
    the even positions we will have normal text and on the odd positions
    we will have the name of a tag.
    In the event that we want to put multiple inference objects in our label
    (for example we want to save the bounding boxes of multiple persons in the same image)
    we will use the separator $$ in order to separate the different fragments of label
    so that we know if a certain sequence of the label will repeat in case of multiple objects.
    Returns
    -------

    """
    # TODO: maybe replace here the aliases from the template file in order to optimize
    if self._template_fragments is None:
      self.P("Init template fragments for dataset builder mixin")
      self._is_multilabel = False
      self._label_keys = set()
      self._mandatory_keys = set(self.get_mandatory_keys)
      template_lines = self.get_label_template_lines()
      lines_str = ''.join(template_lines)
      label_fragments = lines_str.split('$$')
      if len(label_fragments) > 1:
        self._is_multilabel = True
      # endif multiple labels

      self._template_fragments = [
        fragment.split('##')
        for fragment in label_fragments
      ]
      for fragment in self._template_fragments:
        for i, word in enumerate(fragment):
          if i % 2 == 1:
            # odd position in the split line, so we have the name of a tag
            if word.startswith('!'):
              # if the tag starts with '!' it will be considered mandatory (won't be saved if we do not have this value)
              word = word[1:]
              self._mandatory_keys.add(word)
            # endif mandatory key
            self._label_keys.add(word)
      # endfor line in self.template_lines
      for keys_list in [self.get_image_save, self.get_image_crop, self.get_additional_keys]:
        for key in keys_list:
          self._label_keys.add(key)
      # endfor keys_lists
    # endif self._template_fragments is None
    return

  def _maybe_init_path_elements(self):
    """
    Here we will parse the generic path with the purpose of identifying the name tags
    in order to search them both in the inference dictionaries and the payload dictionaries.
    Returns
    -------

    """
    if self._path_elements is None:
      self.P("Init path elements for dataset builder mixin")

      self._path_elements = self.get_generic_path.split('##')
      for i, word in enumerate(self._path_elements):
        if i % 2 == 1:
          # odd position in the split line, so we have the name of a tag
          if word.startswith('!'):
            # if the tag starts with '!' it will be considered mandatory (won't be saved if we do not have this value)
            word = word[1:]
            self._mandatory_keys.add(word)
          # endif mandatory key
          self._label_keys.add(word)
      # endfor i, word in enumerate(self._path_elements)
    # endif self._path_elements is None
    return

  def _init_dataset_builder_params(self):
    # Parameters reset for consistency
    self._ds_params_inferences = []
    self._ds_params_frame = {}
    if self.cfg_dataset_builder is not None:
      self._maybe_init_template_fragments()
      self._maybe_init_path_elements()

    return

  def _maybe_add_original_image(self, dct, img_orig=None):
    """
    Method used to add 'IMG_ORIG' value to the retrieved data dictionary
    in case it is needed and is not already in dct
    Parameters
    ----------
    dct - dictionary of retrieved data
    img_orig - numpy image, image that will be added to the resulted dict if check_img_orig == True
    -if this is not provided the added image will come from self.dataapi_image()

    Returns
    -------

    """
    if dct.get('IMG_ORIG') is None:
      inference_mapping = self.get_inference_mapping
      if 'IMG_ORIG' in self._label_keys or 'IMG_ORIG' in inference_mapping.values():
        dct['IMG_ORIG'] = self.dataapi_image() if img_orig is None else img_orig
      # endif 'IMG_ORIG' necessary
    # endif 'IMG_ORIG' not already in dct
    return dct

  def _extract_data(self, dct, check_img_orig=False, img_orig=None):
    """
    Method for extracting data from either an inference dictionary or a payload dictionary
    with the purpose of further using it for the built dataset.
    Parameters
    ----------
    dct - dictionary of raw information provided either by the inference or by the payload
    check_img_orig - bool, flag that dictates if 'IMG_ORIG' should be in the resulted dictionary
    img_orig - numpy image, image that will be added to the resulted dict if check_img_orig == True
    -if this is not provided the added image will come from self.dataapi_image()

    Returns
    -------
    res - dictionary of the retrieved data
    """
    if dct is None:
      return {}
    res = {
      key: dct.get(
        # in case the current key in self._label_keys is an alias
        # we will try to retrieve the real key from dct
        self.get_inference_mapping.get(key, key)
      )
      for key in self._label_keys
      if self.get_inference_mapping.get(key, key) in dct.keys()
    }
    if 'FULL' in self._label_keys:
      res['FULL'] = 'FULL'
    # endif full image necessary
    if check_img_orig:
      res = self._maybe_add_original_image(res, img_orig=img_orig)
    res[self.consts.TYPE] = dct.get(self.consts.TYPE)
    res[self.consts.TRACK_ID] = dct.get(self.consts.TRACK_ID)
    res[self.consts.APPEARANCES] = dct.get(self.consts.APPEARANCES, 0)

    return res

  def __eligible_object(self, inference):
    """
    Method for determining if the current inference will be saved or not based on the number of appearances.
    Parameters
    ----------
    inference - dictionary describing the current object inference

    Returns
    -------
      True if the inference will be saved
      False otherwise
    """
    track_id = inference.get(self.consts.TRACK_ID)
    if track_id is None or self.get_object_max_saves < 1:
      # the current serving process does not use any tracker, therefore we do not know
      # the current object's identity
      return True
    obj_type = inference.get(self.consts.TYPE)
    if obj_type not in self.__object_saves.keys():
      # first time seeing the current type | also applies if the type will not be provided and obj_type == None
      self.__object_saves[obj_type] = {}
    current_stats = self.__object_saves[obj_type]
    if track_id not in current_stats.keys():
      current_stats[track_id] = 0

    # checking to see if we have a limit for how many times to save
    # an object and if the current object reached that limit
    return current_stats[track_id] < self.get_object_max_saves

  def _maybe_ds_builder_gather_before_process(self, inferences=None):
    """
    Method to process the data from the inferences provided by the serving plugin.

    Parameters
    ----------
    inferences: None or list of inferences
    If None this will take the list of inferences from self.dataapi_image_instance_inferences().
    In case the plugin developer needs to process here a different list of inferences it can
    be done in the process method of the plugin by calling this method and passing the desired
    list of inferences.

    Returns
    -------

    """
    if self.cfg_dataset_builder is None or self.__enough_saves():
      return
    if inferences is None:
      inferences = self.dataapi_image_plugin_inferences()

      if inferences is None:
        return
      # endif secondary check
    # endif inferences provided

    # Reset this in case it is being called for the second time in the plugin
    self._ds_params_inferences = []
    for inf in inferences:
      if self.__eligible_object(inf):
        curr_dict = self._extract_data(inf)
        self._ds_params_inferences.append(curr_dict)
    # endfor inf in inferences
    return

  def __maybe_gather_from_payload(self, payload):
    """
    Method for gathering data from a certain payload if it was not used before
    Parameters
    ----------
    payload

    Returns
    -------

    """
    if not vars(payload).get('_P_DATASET_BUILDER_USED', False) and not vars(payload).get('IS_NO_DATA_ALERT', False):
      self._ds_params_frame = self._extract_data(vars(payload), check_img_orig=True)
      vars(payload)['_P_DATASET_BUILDER_USED'] = True
    # endif unused payload
    return

  def _maybe_ds_builder_gather_after_process(self, payload=None):
    """
    # TODO: discuss strategy for multiple payloads
    Method to process the data from the provided payload.
    If not payload is provided the last payload provided by the process method will be used.

    Parameters
    ----------
    payload - GeneralPayload, result of an iteration of a certain plugin

    Returns
    -------

    """
    if self.cfg_dataset_builder is None or self.__enough_saves():
      return

    if payload is None:
      # This is done in order to keep the values of the last payload
      # but in case the last payload does not have a specific field
      # we also search it in the previous available payloads
      for payload in self.payloads_deque:
        self.__maybe_gather_from_payload(payload=payload)
      # endfor payloads

      if self._payload is not None:
        self.__maybe_gather_from_payload(payload=self._payload)
    else:
      self.__maybe_gather_from_payload(payload=payload)
    # endif payload provided
    return

  def _process_saved_data(self, data):
    """
    Method present in order to allow the user to perform custom actions on the data
    that will be saved for our current dataset.
    Parameters
    ----------
    data - dictionary of data from both model serving and payload

    Returns
    -------
    res - dictionary with the processed data
    """
    return data

  def check_mandatory_keys(self, data, mandatory_keys=None):
    """
    Method used for checking if a certain dictionary has all the mandatory keys or not
    Parameters
    ----------
    data - dictionary
    mandatory_keys - list of mandatory keys

    Returns
    -------

    """
    if mandatory_keys is None:
      mandatory_keys = self._mandatory_keys
    # endif provided mandatory_keys
    for key in mandatory_keys:
      if data.get(key) is None:
        return False
      # endif mandatory key present in data dictionary
    # endfor mandatory keys
    return True

  def _process_saved_data_multilabel(self, data, inferences):
    """
    Method present in order to allow the user to perform custom actions on the data
    that will be saved for our current dataset in case of multiple inferences in the same label.
    By default, this method will apply _process_saved_data on each inference and then merge the results.
    Parameters
    ----------
    data - dictionary of data from payload
    inferences - list of inferences from which we will extract the data for the current fragment

    Returns
    -------
    res - (new_data, new_inferences), where:
    new_data - dictionary with the processed data from the payload
    new_inferences - list of dictionaries with the processed data from the inferences
    """
    processed_data_list = [self._process_saved_data({**data, **inf}) for inf in inferences]
    new_inferences = []
    new_data = data
    for i, processed_data in enumerate(processed_data_list):
      new_inference = {
        **processed_data,
        **inferences[i]
      }
      new_data = {
        key: processed_data.get(key, val)
        for key, val in data.items()
      }
      new_inferences.append(new_inference)
    # endfor processed_data_list
    return new_data, new_inferences

  def __process_template_fragment(self, fragment, data):
    """
    Method for converting fragment of template file in a fragment for specific data point.
    Parameters
    ----------
    fragment - list of strings that define a fragment from the template file
    The strings from fragment will be as follows:
    fragment[2 * k] - normal text from the template file
    fragment[2 * k + 1] - tag describing text from the template file
    data - dictionary of data from both model serving and payload

    Returns
    -------
      res - concatenation of the converted strings contained in line
    """
    return ''.join([
      str(
        data.get(
          word if not word.startswith('!') else word[1:],  # in case our current word is an alias we will decode it
          # using the mapping provided in the "INFERENCE_MAPPING" parameter of the dataset builder
          '_'  # in case we do not have the desired data in our dictionary we will print _ instead
        )
      ) if i % 2 == 1 else word
      for i, word in enumerate(fragment)
    ])

  def __process_repeatable_template_fragment(self, fragment, data, inferences):
    """
    Method for converting fragment of template file in a fragment for specific data point
    in case we will have to repeat the fragment for multiple inferences.
    Parameters
    ----------
    fragment - list of strings that define a fragment from the template file
    The strings from fragment will be as follows:
    fragment[2 * k] - normal text from the template file
    fragment[2 * k + 1] - tag describing text from the template file
    data - dictionary of data from both model serving and payload
    inferences - list of inferences from which we will extract the data for the current fragment

    Returns
    -------
      res - concatenation of the converted strings contained in line
    """
    return ''.join([
      self.__process_template_fragment(
        fragment=fragment,
        data={
          **data,
          **inf
        }
      )
      for inf in inferences
    ])

  def _generate_label_file_text(self, data, inferences=None):
    """
    Method for generating the content in the label file based on the text lines
    in the template file and the specific data for the current datapoint.
    Parameters
    ----------
    data - dictionary of data from both model serving and payload
    inferences - list of inferences from which we will extract the data
    for the current fragment in case of multiple inferences in the same label

    Returns
    -------
      res - concatenation of the processed template lines using specific values
    """
    res = ''.join([
      self.__process_template_fragment(fragment, data)
      if i % 2 == 0 else
      self.__process_repeatable_template_fragment(fragment, data, inferences)
      for i, fragment in enumerate(self._template_fragments)
    ])
    return res

  def _compute_saved_x(self, source, used_prop):
    """
    Method used for computing the saved data used for the X files.
    By default, source will have to be an image and used_prop either
    the string 'FULL' or an array of 4 values that will be considered
    the top, left, bottom, right coordinates of the desired crop from the source.
    The plugin user can customize this however he needs based on its plugin.
    Parameters
    ----------
    source - source of data
    used_prop - property used to extract data from the source

    Returns
    -------
    saved_data - data that will be saved
    """
    if used_prop == 'FULL':
      return source
    t, l, b, r = used_prop
    expand = self.get_expand_value
    if expand > 0:
      if expand > 1:
        expand = int(expand)
        t, l, b, r = max(t - expand, 0), max(l - expand, 0), b + expand, r + expand
      else:
        h = b - t
        w = r - l
        exp_h = int(self.np.ceil(h * expand))
        exp_w = int(self.np.ceil(w * expand))
        t, l, b, r = max(t - exp_h, 0), max(l - exp_w, 0), b + exp_h, r + exp_w
    return source[t: b, l: r]

  def _save_x_method(self, data_dict, subdir, fn, source_key, prop_key):
    """
    Method used for saving data for X file based on its source and the used property.
    Can be customised by the plugin developer.
    Parameters
    ----------
    data_dict - dictionary of data from both model serving and payload
    fn - name of the saved file
    source_key - key from data_dict indicating value used as data source
    prop_key - key from data_dict indicating value used for extracting data from the source
    Returns
    -------
    True if saved, False otherwise
    """
    source = data_dict.get(source_key)
    used_prop = data_dict.get(prop_key)
    if source is not None and used_prop is not None:
      try:
        saved_data = self._compute_saved_x(source=source, used_prop=used_prop)
        self.diskapi_save_image_output(
          image=saved_data,
          filename=fn,
          subdir=subdir,
          extension=self.get_dataset_builder_params.get(self.ds_consts.X_EXTENSION, '')
        )
      except Exception as e:
        return False
    # endif found source and used_prop
    return True

  """PATH HELPERS SECTION"""
  if True:
    def _generate_file_name(self, data):
      """
      Method for generating the file name of the current datapoint.
      Parameters
      ----------
      data - dictionary of data from both model serving and payload

      Returns
      -------
      fn - filename used for the label and input files
      """
      track_id = data.get(self.consts.TRACK_ID)
      obj_type = data.get(self.consts.TYPE)
      if track_id is None:
        current_fn = f'{"__".join(self.unique_identification)}_{self.__saved_data_counter}'
      else:
        # the following checks should not be necessary but are here for safety reasons
        if obj_type not in self.__object_saves.keys():
          self.__object_saves[obj_type] = {}
        curr_saves = self.__object_saves[obj_type]
        if track_id not in curr_saves.keys():
          curr_saves[track_id] = 0

        n_saves = curr_saves[track_id]
        type_prefix = '' if obj_type is None else f'_{obj_type}'
        current_fn = f'{"__".join(self.unique_identification)}{type_prefix}_{track_id}_{n_saves}_{self.__saved_data_counter}'
        curr_saves[track_id] += 1
      # endif track_id is None

      return current_fn

    def parse_generic_path(self, data):
      """
      Method used to generate the path where the current datapoint will be saved at.
      Parameters
      ----------
      data - dictionary of data from both model serving and payload

      Returns
      -------

      """
      return ''.join([
        word if i % 2 == 0 else str(data.get(word, 'unknown'))
        for i, word in enumerate(self._path_elements)
      ])

    def _generate_subdir(self, data):
      """
      Method for generating datapoint subdirectory in case needed
      Parameters
      ----------
      data - dictionary of data from both model serving and payload

      Returns
      -------
      subdir - subdirectory in the dataset
      """
      return self.parse_generic_path(data=data)

    def __get_ds_dir_path(self):
      """
      Private method for accessing the name of the saving directory.
      Returns
      -------
      dir - name of the directory where all the data will be found.
      """
      return self.os_path.join('dataset', self.get_dataset_name())

    def __get_current_chunk_path(self):
      """
      Private method for accessing the name of the current saving directory(this is used for zipping
      files without worrying about the loss of files saved while the archived files are deleted).
      Returns
      -------
      dir - name of the directory where the current chunk of data will be found.
      """
      return f'{self.get_dataset_name()}_{self.__dir_cnt}'

    def __get_subdir_path(self, data):
      """
      Private method for generating the full path of the current datapoint's saved files
      Parameters
      ----------
      data - dictionary of data from both model serving and payload

      Returns
      -------

      """
      custom_subdir = str(self._generate_subdir(data=data))
      subdir = self.os_path.join(self.__get_ds_dir_path(), self.__get_current_chunk_path())
      if len(custom_subdir) > 0:
        subdir = self.os_path.join(subdir, custom_subdir)

      return subdir

    def get_saved_stats(self):
      """
      Method used for retrieving the saved stats of the dataset builder.
      Returns
      -------
      stats_dict - dictionary containing the saved stats
      """
      stats_dict = self.diskapi_load_pickle_from_output(
        filename='saved_stats.pkl',
        subfolder=self.__get_ds_dir_path(),
      )
      return stats_dict

    def update_saved_stats(self):
      """
      Method used for updating the saved stats of the dataset builder.
      Returns
      -------

      """
      self.P(
        f'Updating saved stats. Current saved datapoints: {self.__saved_data_counter} and current chunk: {self.__dir_cnt}'
      )
      stats_dict = {
        'saved_data_counter': self.__saved_data_counter,
        'dir_cnt': self.__dir_cnt,
        'object_saves': self.__object_saves
      }
      self.diskapi_save_pickle_to_output(
        obj=stats_dict,
        filename='saved_stats.pkl',
        subfolder=self.__get_ds_dir_path(),
      )
      return

    def _archive_and_upload_current_chunk(self):
      """
      Method for archiving and uploading the current chunk of data.
      Returns
      -------

      """
      curr_dir_name = self.__get_current_chunk_path()
      self.P(
        f'Archiving and uploading current data chunk ({curr_dir_name})[containing {(self.__saved_data_counter - 1) % self.get_zip_period + 1} datapoints]')
      curr_zip_name = f'{self.__first_timestamp}_{self.__get_current_chunk_path()}.zip'
      save_path = self.os_path.join(self.log.get_target_folder(target='output'), self.__get_ds_dir_path())
      dir_path = self.os_path.join(save_path, curr_dir_name)
      zip_path = self.os_path.join(save_path, curr_zip_name)

      self.diskapi_zip_dir(dir_path=dir_path, zip_path=zip_path)
      self.diskapi_delete_directory(dir_path=dir_path)
      self.upload_file(
        file_path=zip_path,
        target_path=self.os_path.join(
          self.get_dataset_name(),
          curr_zip_name
        ),
        bucket_name='ds-builder-datasets',
        verbose=1
      )
      return
  # endif
  """END PATH HELPERS SECTION"""

  def _save_datapoint(self, data, inferences=None):
    """
    Method for saving a single datapoint from the gathered data.
    Parameters
    ----------
    data - dictionary of data from both model serving and payload

    Returns
    -------
    True if saved, False otherwise
    """
    success = False
    label_success = False
    current_fn = self._generate_file_name(data=data)
    current_subdir = self.__get_subdir_path(data=data)
    if self.get_label_template_type is not None:
      label_file_text = self._generate_label_file_text(data, inferences=inferences)
      label_success = self.diskapi_save_file_to_output(
        data=label_file_text,
        filename=current_fn,
        subdir=current_subdir,
        extension=self.get_dataset_builder_params.get(self.ds_consts.LABEL_EXTENSION, '')
      )
    # endif save label_file
    x_file_params = self.get_x_file_params()
    for x_file_param in x_file_params:
      source_key, prop_key = x_file_param
      fn_suffix = '' if len(x_file_params) < 2 else f'_{source_key}_{prop_key}'
      save_result = self._save_x_method(
        data_dict=data,
        subdir=current_subdir,
        fn=current_fn + fn_suffix,
        source_key=source_key,
        prop_key=prop_key
      )
      if save_result:
        success = True
      # endif save successful
    # endfor x_file_params
    success = success or label_success
    if success:
      self.__saved_data_counter += 1
      if self.get_zip_period > 0 and self.__saved_data_counter % self.get_zip_period == 0:
        self._archive_and_upload_current_chunk()
        self.P(f'Incrementing current chunk from {self.__dir_cnt} to {self.__dir_cnt + 1}')
        self.__dir_cnt += 1
      # endif archive current chunk
      if self.__saved_data_counter % self.get_stats_update_period == 0:
        self.update_saved_stats()
        self.P(f'Updated saved stats. Current saved datapoints: {self.__saved_data_counter}')
    # endif success
    return success

  def maybe_archive_upload_last_files(self):
    """
    Method used for archiving and uploading the remaining datapoints (if it's the case) when the plugin instance closes.
    Returns
    -------

    """
    if self.cfg_dataset_builder is None:
      return
    # endif dataset builder configured
    if self.get_zip_period > 0 and self.__saved_data_counter % self.get_zip_period > 0:
      self._archive_and_upload_current_chunk()
      self.__dir_cnt += 1
      self.update_saved_stats()
    # endif unarchived datapoints
    return

  def is_valid_datapoint(self, data):
    """
    Method used for checking if the provided data is valid for saving or not.
    This is implemented in order for the plugin developer to be able to custom filter
    the data that will be saved.
    Parameters
    ----------
    data

    Returns
    -------
    True if valid, False otherwise
    """
    return True

  def __is_valid_datapoint(self, data):
    """
    Wrapper method used for checking if the provided data is valid for saving or not.
    Parameters
    ----------
    data - dictionary of data from both model serving and payload

    Returns
    -------
    True if valid, False otherwise
    """
    return self.check_mandatory_keys(data=data) and self.is_valid_datapoint(data)

  def _maybe_save_datapoint(self, data):
    """
    Wrapper method of _save_datapoint that also checks for the
    data dictionary to have all the mandatory keys
    Parameters
    ----------
    data - dictionary of data from both model serving and payload

    Returns
    -------
    True if saved, False otherwise
    """
    try:
      processed_data = self._process_saved_data(self.deepcopy(data))
      return self.__is_valid_datapoint(data=processed_data) and self._save_datapoint(data=processed_data)
    except:
      return False

  def filter_inferences_idx(self, data, inferences):
    """
    Method used for filtering the inferences that will be saved.
    This is implemented in order for the plugin developer to be able to custom filter
    the data that will be saved.
    Parameters
    ----------
    inferences - list of inferences

    Returns
    -------
    res - list of filtered inferences indexes
    """
    return [
      i for i, inf in enumerate(inferences)
      if self.__is_valid_datapoint(data={**data, **inf})
    ]

  def filter_inferences(self, data, inferences):
    """
    Method used for filtering the inferences that will be saved.
    This is implemented in order for the plugin developer to be able to custom filter
    the data that will be saved.
    Parameters
    ----------
    inferences - list of inferences

    Returns
    -------
    res - list of filtered inferences
    """
    return [inferences[i] for i in self.filter_inferences_idx(data=data, inferences=inferences)]

  def _maybe_save_multilabel_datapoint(self, data, inferences):
    """
    Wrapper method of _save_datapoint that is used in case of multiple inferences in the same label.
    Parameters
    ----------
    data - dictionary of data from both model serving and payload
    inferences - list of inferences from which we will extract the data for the current fragment

    Returns
    -------
    res = (success, filtered)
    success - True if saved something, False otherwise
    filtered - list of inferences that were saved
    """
    try:
      # TODO: maybe refactor this so that the user will receive the filtered inferences in the process method
      # maybe have 2 methods of processing the data, one before the validation and one after
      proc_data, proc_inferences = self._process_saved_data_multilabel(data, inferences)
      filtered_idxs = self.filter_inferences_idx(data=proc_data, inferences=proc_inferences)
      filtered_proc_inferences = [proc_inferences[i] for i in filtered_idxs]
      filtered_inferences = [inferences[i] for i in filtered_idxs]
      if len(filtered_idxs) == 0:
        # if no inferences were kept after filtering it means we have a background image
        if self.get_background_period_save == 0 or self._background_saves % self.get_background_period_save > 0:
          return False, filtered_inferences
        # endif background period
      # endif no inferences
      success = self._save_datapoint(data=self.deepcopy(proc_data), inferences=filtered_proc_inferences)
      if len(filtered_idxs) == 0:
        self._background_saves += success
      # endif background saved
      return success, filtered_inferences
    except:
      return False, []

  def _maybe_ds_builder_save(self):
    """
    # TODO: add save every x seconds
    # TODO: add maybe_upload at the end of this method
    Main method of the mixin.
    Here we save the gathered data from both the model serving and the payload generated by the plugin.
    Returns
    -------

    """
    if self.cfg_dataset_builder is None or self.__enough_saves():
      return

    # check to see if enough time has passed from the last save
    if self.get_save_cap_resolution() > 0 and self._last_save_ts is not None \
            and self.time() - self._last_save_ts < 1 / self.get_save_cap_resolution():
      return
    # endif
    success = False
    if len(self._ds_params_frame.keys()) < 1:
      self._ds_params_frame = self._extract_data(self._ds_params_frame, check_img_orig=True)
    # endif no frame data
    if self._is_multilabel:
      inferences = [inf for inf in self._ds_params_inferences if not inf.get('SAVED', False)]
      current_params = self._ds_params_frame
      success, filtered_inferences = self._maybe_save_multilabel_datapoint(data=current_params, inferences=inferences)
      for inf in filtered_inferences:
        inf['SAVED'] = success
      # endfor filtered_inferences
    else:
      for inference_params in self._ds_params_inferences:
        if inference_params.get('SAVED', False):
          continue
        # endif datapoint already saved
        current_params = {
          **self._ds_params_frame,
          **inference_params
        }
        # if self.consts.TRACK_ID in inference_params and self.consts.TRACK_ID not in current_params.keys():
        #   current_params[self.consts.TRACK_ID] = inference_params[self.consts.TRACK_ID]
        #   current_params[self.consts.APPEARANCES] = inference_params[self.consts.APPEARANCES]

        inference_params['SAVED'] = self._maybe_save_datapoint(current_params)
        if inference_params['SAVED']:
          success = True
        # endif success
      # endfor inferences data
    # endif multilabel
    if success:
      self._last_save_ts = self.time()
    # endif success
    return



from naeural_core import DecentrAIObject
from decentra_vision.draw_utils import DrawUtils


class AbstractModel(DecentrAIObject):
  """
  As the name indicates, this is an abstract class that will impose a template
  for future model benchmarks. 
  Each benchmarked model should inherit this class.
  """

  def __init__(self, **kwargs):
    super().__init__(**kwargs)
    self.init(**kwargs)

    self._painter = DrawUtils(log=self.log)
    return

  def init(self, **kwargs):
    """
    Method used to initialize variables/methods that your model needs.
    """
    raise NotImplementedError
    return

  def load(self, **kwargs):
    """
    Loads model/graph but does not executes GPU operations like sessions or .to(CUDA)
    """
    raise NotImplementedError
    return

  def prepare(self, **kwargs):
    """
    This method will create GPU specific ops like session creation, move model to cuda (in torch), etc.

    """
    raise NotImplementedError
    return

  def _predict(self, inputs, **kwargs):
    """
    Specific method used to predict

    """
    return

  def _process_input(self, inputs, **kwargs):
    """
    Specific method used to preprocess input

    """
    raise NotImplementedError
    return

  def draw(self, path_out, inputs, preds):
    raise NotImplementedError
    return

  def predict(self, inputs, **kwargs):
    self.log.start_timer('predict')

    inputs = self._process_input(inputs=inputs, **kwargs)
    preds = self._predict(inputs)

    self.log.stop_timer('predict')
    return preds



import uuid 

from threading import Lock

class Singleton:
  _lock: Lock = Lock()
  def __new__(cls, log, **kwargs):
    with cls._lock:
      if not hasattr(cls, '_instance'):
        instance = super(Singleton, cls).__new__(cls)
        instance.uuid = str(uuid.uuid4())[:4]
        instance.log = log
        instance.build(**kwargs)
        cls._instance = instance
      else:
        instance = cls._instance
    return instance
  
  def build(self, **kwargs):
    raise NotImplementedError('This method must be implemented by the subclass')
  

class Multitone:
  _lock: Lock = Lock()
  def __new__(cls, name, log, **kwargs):
    with cls._lock:
      if not hasattr(cls, '_instances'):
        cls._instances = {}
      if name not in cls._instances:
        instance = super(Multitone, cls).__new__(cls)
        instance.uuid = str(uuid.uuid4())[:8]
        instance.log = log
        instance.name = name
        instance.build(**kwargs)
        cls._instances[name] = instance
      else:
        instance = cls._instances[name]
    return instance
  
  def build(self, **kwargs):
    raise NotImplementedError('This method must be implemented by the subclass')
  
  def __del__(self):
    with self._lock:
      if hasattr(self, 'name') and self.name in self.__instances:
        del self.__instances[self.name]
    return
  
  
if __name__ == '__main__':
  log = "LOG"
  
  class TestSingleton(Singleton):
    def build(self, param1=None, param2=None):
      print('Building TestSingleton with params: {} {}'.format(param1, param2))
      return
    
  class TestMultitone(Multitone):
    def build(self, p1=1, p2=2, p3=3):
      print('Building TestMultitone {} with params: {} {}'.format(self.name, p1, p2, p3))
      return
    
    
  class TestContainerBase:
    def __init__(self, log=None):
      self.log = log
      self.startup()
      return
    
    def startup(self):
      raise NotImplementedError('This method must be implemented by the subclass')
  
  class TestContainer(TestContainerBase):
    def __init__(self, log=None):
      self.s1 = None
      super().__init__(log=log)
      return
    
    def startup(self):
      self.s1 = TestSingleton(log=self.log, param1='a', param2='b')
      return
    
    
    
  s1 = TestSingleton(log=log, param1='a', param2='b')
  s2 = TestSingleton(log=log, param1='c', param2='d')
  
  m1 = TestMultitone(name='m1', log=log, p1=1, p2=2, p3=3)
  m2 = TestMultitone(name='m2', log=log, p1=4, p2=5, p3=6)
  m3 = TestMultitone(name='m1', log=log, p1=7, p2=8, p3=9)
  m4 = TestMultitone(name='m2', log=log, p1=10, p2=11, p3=12)
  
  c1 = TestContainer(log=log)
  
  assert s1.uuid == s2.uuid
  assert m1.uuid == m3.uuid
  assert m2.uuid == m4.uuid
  
  assert s1.uuid == c1.s1.uuid
  


import os

class _KerasCallbacksMixin(object):
  """
  Mixin for keras callbacks functionalities that are attached to `libraries.logger.Logger`.

  This mixin cannot be instantiated because it is built just to provide some additional
  functionalities for `libraries.logger.Logger`

  In this mixin we can use any attribute/method of the Logger.

  * Obs: This mixin uses also attributes/methods of `_BasicTFKerasMixin`:
    - self.TF
  """

  def __init__(self):
    super(_KerasCallbacksMixin, self).__init__()
    try:
      from .basic_tfkeras_mixin import _BasicTFKerasMixin
    except ModuleNotFoundError:
      raise ModuleNotFoundError("Cannot use _KerasCallbacksMixin without having _BasicTFKerasMixin")

    self._tensorboard_dir = None
    return

  def get_keras_lr_callback(self, monitor='loss', patience=3, factor=0.1,
                            use_tf_keras=True):
    import tensorflow as tf
    if use_tf_keras:
      assert self.TF
      cb = tf.keras.callbacks.ReduceLROnPlateau(monitor=monitor, patience=patience, factor=factor)
    else:
      # assert self.KERAS
      from keras.callbacks import ReduceLROnPlateau
      cb = ReduceLROnPlateau(monitor=monitor, patience=patience, factor=factor)
    return cb

  def get_keras_checkpoint_callback(self, model_name='model',
                                    monitor='loss', mode='auto',
                                    period=1, use_tf_keras=True, save_best_only=False):
    import tensorflow as tf
    dict_abbrv = {'val_loss': 'ValL', 'loss': 'L', 'K_rec': 'RECA',
                  'K_pre': 'PREC', 'K_f2': 'F2SC',
                  'K_r2': 'RSQ', 'val_K_pre': 'ValP', 'val_K_rec': 'ValR'}
    try:
      abbrv = dict_abbrv[monitor]
    except:
      abbrv = monitor[:3]

    if save_best_only:
      file_path = model_name + '_best_{}'.format(monitor) + ".h5"
    else:
      file_path = model_name + '_tk_E{epoch:02d}_' + abbrv + "{" + monitor + ":.6f}" + ".h5"

    file_path = self.file_prefix + '_' + file_path

    if use_tf_keras:
      assert self.TF

      self.verbose_log("Creating tf.keras chekpoint callback...")
      file_path = os.path.join(self.get_models_folder(), file_path)
      cb = tf.keras.callbacks.ModelCheckpoint(
        filepath=file_path,
        monitor=monitor,
        verbose=0,
        save_best_only=save_best_only,
        save_weights_only=False,
        mode=mode,
        period=period
      )
    else:
      # assert self.KERAS
      from keras.callbacks import ModelCheckpoint
      self.verbose_log("Creating keras chekpoint callback...")
      file_path = os.path.join(self.get_models_folder(), file_path)
      cb = ModelCheckpoint(
        filepath=file_path,
        monitor=monitor,
        verbose=0,
        save_best_only=save_best_only,
        save_weights_only=False,
        mode=mode,
        period=period
      )

    return cb

  def get_keras_tensorboard_callback(self, use_tf_keras=True):
    """
    """
    import tensorflow as tf
    if use_tf_keras:
      assert self.TF == True

      self.verbose_log("Creating tf.keras tensorboard callback...")
      self._tensorboard_dir = os.path.join(self.get_base_folder(), '_tf');
      if not os.path.isdir(self._tensorboard_dir):
        os.makedirs(self._tensorboard_dir)
      cb_tboard = tf.keras.callbacks.TensorBoard(
        log_dir=self._tensorboard_dir,
        histogram_freq=1,
        write_graph=True,
        write_images=True
      )
    else:
      from keras.callbacks import TensorBoard
      # assert self.KERAS == True
      self.verbose_log("Creating Keras tensorboard callback...")
      self._tensorboard_dir = os.path.join(self.get_base_folder(), '_tf');
      if not os.path.isdir(self._tensorboard_dir):
        os.makedirs(self._tensorboard_dir)
      cb_tboard = TensorBoard(
        log_dir=self._tensorboard_dir,
        histogram_freq=1,
        write_graph=True,
        write_images=True
      )
    return cb_tboard




from naeural_core.main.ver import __VER__

if __name__ == "__main__":
  with open("pyproject.toml", "rt", encoding="utf-8") as fd:
    new_lines = []
    lines = fd.readlines()
    for line in lines:
      if "version" in line:
        line = f'version = "{__VER__}"\n'
      new_lines.append(line)

  with open("pyproject.toml", "wt", encoding="utf-8") as fd:
    fd.writelines(new_lines)



import os
import json
import traceback
import numpy as np

from time import time as tm
from collections import OrderedDict
from datetime import datetime as dt

from .advanced_callback import AdvancedCallback


def get_adv_callback(log,
                     model_name=None,
                     lr_monitor='val_loss',
                     lr_mode='min',
                     lr_factor=0.1,
                     lr_patience=2,
                     lr_min_delta=1e-4,
                     lr_min_lr=1e-7,
                     save_monitor='val_loss',
                     save_mode='min',
                     save_last=2,
                     save_weights=False,
                     y_trans=None,
                     y_scale=None,
                     metric_calc=None,
                     calc_MAPE_TS=False,
                     calc_MAD_TS=None,
                     allow_neg_test=False,
                     log_validate_each=1,
                     DEBUG=False,
                     ):
  """
  model_name : if None the model name will be inferred from tf.keras.models.Model.name
  lr_* : parameters for learning rate monitor

  save_monitor : what should be measured
  save_mode : how the checkpoint monitor should be measured
  save_last : the monitor will only keep best 'save_last' files

  metric_calc: function with signature func(model, data1, data2) -> dict
               with "metric":value pairs

  y_scale, y_tras: transform y if needed

  log_validate_each :  will log chosen validation metrics every x epochs (default 1)

  calc_MAPE_TS: (default None) used in TS prediction calcs MAPE on last
                time-step on validation data
  calc_MAD_TS: (default None) value x used in TS prediction to calculate
                "period average deviation" for last x steps period


  *****

  `log` :  obiectul Logger din care se apeleaza - este automat dat de Logger - dont bother
  `model_name=None` : optional numele modelului urmarit - va lua numele Keras if None
  `lr_monitor='val_loss'` : ce monitorizam pentru scaderea ratei de invatare
  `lr_mode='min'` : cum monitorizam - min inseamna ca urmarim minimizarea lui lr_monitor
  `lr_factor=0.1` : cu ce scalam rata daca nu se respecta lr_mode pentru lr_patience pasi
  `lr_patience=3` : cati pasi avem 'rabdare' cu rata curenta
  `lr_min_delta=1e-4` : care este minimum de scadere/crestere a lr_monitor
  `lr_min_lr=1e-7`  : pana unde scadem de mult rata
  `save_monitor='val_loss'` : ce monitorizam ca sa salvam cele mai bune epoci
  `save_mode='min'` : cum monitorizam ...
  `save_last=2` : top epoci salvate
  `metric_calc=None` :  functie cu semnatura f(model,x_val,y_val)->dict(metric:val)
  `calc_MAPE_TS=False` : calcularea val_MAPE pe ultimul pas al series - doar pentru TS
  `calc_MAD_TS=None` : calculeaza val_MAD pe ultimii `calc_MAD_TS` pasi din TS
  `log_validate_each=1` : printeaza date validare la fiecare nr de epoci
  `y_trans=None` : variabila de translatare a output-ului de validare (daca a fost transf inainte)
  `y_scale=None` : variabile de scalare --//----
  `DEBUG=False` : True ca sa arate informatii extra la fiecare epoca
  `allow_neg_test=False` : Negative va face pozitive toate valorile pentru teste
  `save_weights=False` : true ca sa salveze doar weights




  Exemple:

  `model.fit(...., validation_data=(X_val,Y_val),
             callbacks=[log.get_adv_callback(calc_MAPE_TS=True, lr_monitor=None)`
                          va calcula folosind datele de validare val_MAPE dupa
                          fiecare epoca si nu va umbla deloc la rata de invatare
  `model.fit(...., validation_data=(X_val,Y_val),
             callbacks=[log.get_adv_callback(calc_MAD_TS=5)`
             va calcula folosind datele de validare `val_MAD` pe ultimii 5
             pasi ai fiecarei serie dupa fiecare epoca si utiliza `val_loss` (default)
             ca sa modifice rata de invatare daca pare sa nu convearga.
  """

  cb = AdvancedCallback(log=log,
                        model_name=model_name,
                        lr_monitor=lr_monitor,
                        lr_mode=lr_mode,
                        lr_factor=lr_factor,
                        lr_patience=lr_patience,
                        lr_min_delta=lr_min_delta,
                        lr_min_lr=lr_min_lr,
                        save_monitor=save_monitor,
                        save_mode=save_mode,
                        y_scale=y_scale,
                        y_trans=y_trans,
                        save_last=save_last,
                        metric_calc=metric_calc,
                        calc_MAPE_TS=calc_MAPE_TS,
                        calc_MAD_TS=calc_MAD_TS,
                        log_validate_each=log_validate_each,
                        DEBUG=DEBUG,
                        save_weights=save_weights,
                        )
  return cb

def retrofit_faruqui(
        log,
        np_X,
        dct_edges,
        max_iters=100,
        tol=5e-3,
        alpha=None,
        beta=None):
  """
  Implements retrofitting method of Faruqui et al.

  Inputs:
  ======
  np_X : np.ndarray
    This is the input embedding matrix

  dct_edges: dict
    This is the dict that maps a certain vector to all its relatives

  max_iters: int (default=100)

  alpha, beta: callbacks that return floats as per paper alpha/beta

  tol : float (default=1e-2)
    If the average distance change between two rounds is at or
    below this value, we stop. Default to 10^-2 as suggested
    in the paper.


  Outputs:
  ======
    np.ndarray: the retrofitted version of np_X

  """

  if alpha is None:
    alpha = lambda x: 1.0
  if beta is None:
    beta = lambda x: 1.0 / len(dct_edges[x])

  np_Y = np_X.copy()
  np_Y_prev = np_Y.copy()
  log.start_timer('retrofit_process')
  log.P("Performing retrofitting on {} embedding matrix...".format(np_X.shape))
  for iteration in range(1, max_iters + 1):
    for i, vec in enumerate(np_X):
      neighbors = dct_edges[i]
      n_neighbors = len(neighbors)
      if n_neighbors:
        a = alpha(i)
        b = beta(i)
        retro = np.array([b * np_Y[j] for j in neighbors])
        retro = retro.sum(axis=0) + (a * np_X[i])
        norm = np.array([b for j in neighbors])
        norm = norm.sum(axis=0) + a
        np_Y[i] = retro / norm
      # end if
    # end for matrix rows
    changes = log._measure_changes(np_Y, np_Y_prev)
    if changes <= tol:
      log.P("Retrofiting converged at iteration {}; change was {:.4f} ".format(
        iteration, changes))
      break
    else:
      np_Y_prev = np_Y.copy()
      print("\rIteration {:d}; change was {:.4f}".format(iteration, changes),
            flush=True,
            end='')
  # end for iterations
  log.end_timer('retrofit_process')
  return np_Y

def retrofit_vector_to_embeddigs(log, np_start_vect, np_similar_vectors):
  """
  this function will retrofit a single vector (can be even a random vector but
  preferably a centroid from the original latent space) to a pre-prepared matrix
  of similar embeddings using the basic Faruqui et al approach
  """
  log.P("Creating new item embedding starting from a {} vector and {} similar items".format(
    np_start_vect.shape, np_similar_vectors.shape))
  log.P("  Current distance: {:.2f}".format(
    log._measure_changes(np_start_vect, np_similar_vectors)))
  n_simils = np_similar_vectors.shape[0]
  np_full = np.concatenate((np_start_vect.reshape(-1, 1),
                            np_similar_vectors))
  dct_edges = {0: [x for x in range(1, n_simils)]}
  np_new_embeds = log.retrofit_faruqui(np_full, dct_edges)
  np_new_embed = np_new_embeds[0]
  log.P("  New distance: {:.2f}".format(
    log._measure_changes(np_new_embed, np_similar_vectors)))
  return np_new_embed



def save_tf_graph(log, tf_saver, tf_session, file_name, sub_folder='', debug=False):
  if file_name[-5] != '.ckpt':
    file_name += '.ckpt'
  mfolder = log.get_models_folder()
  folder = os.path.join(mfolder, sub_folder)
  if not os.path.isdir(folder):
    log.P("Creating folder [{}]".format(folder))
    os.makedirs(folder)
  path = os.path.join(folder, file_name)
  try:
    if debug:
      log.P("Saving tf checkpoint '{}'".format(file_name))
    tf_saver.save(tf_session, path)
  except:
    log.P("ERROR Saving session for {}".format(path[-40:]))
  return

def save_tf_graph_checkpoint(log, session, placeholders, operations, epoch, save_path):
  '''
    Saves a tensorflow graph, given placeholders and operations from it's session to save_path
    Params:
      session     : tf.Session: Current session of the graph to save
      placeholders: Array: tf.placeholder
      operations  : Array: tf.operation, tf.tensor (operations and metrics)
      epoch       : Int: epoch at which the checkpoint is made
      save_path   : String: log-explanatory
    Returns:
      string: save path returned by saver.save
  '''
  import tensorflow.compat.v1 as tf

  with session.graph.as_default():
    log.P("Saving tf graph in [..{}] ...".format(save_path[-70:]))
    saver = tf.train.Saver()

    placeholders_collection = tf.get_collection("Placeholders")
    operations_collection = tf.get_collection("Operations")

    for p in placeholders:
      if p not in placeholders_collection:
        tf.add_to_collection("Placeholders", p)
      # endif
    # endfor

    for o in operations:
      if o not in operations_collection:
        tf.add_to_collection("Operations", o)
      # endif
    # endfor

    ret = saver.save(session, save_path, global_step=epoch)
    log.P("  ", show_time=True)
    return ret
  # endwith

def load_tf_graph_checkpoint(log, session, epoch, save_path):
  '''
    Loads a tensorflow graph values into the current given session
    Params:
      session     : tf.Session: Current session of the graph in which to load
      placeholders: Array: labels given to placeholders which will be fetched from the saved model
      operations  : Array: labels given to operations and metrics which will be fetched from the saved model
      epoch       : Int: epoch at which the checkpoint was made
      save_path   : String: log-explanatory
    Returns:
      Tuple containing a dictionary with the placeholders and a dictionary with the operations
  '''
  import tensorflow.compat.v1 as tf
  saver = tf.train.import_meta_graph(save_path + '-' + str(epoch) + '.meta')
  saver.restore(session, save_path + '-' + str(epoch))

  placeholders_dict = tf.get_collection("Placeholders")
  operations_dict = tf.get_collection("Operations")

  return (placeholders_dict, operations_dict)

###
### Keras-to-TF section and Keras inference helpers
###


def model_h5_to_graph(log, file_name,
                      custom_objects=None,
                      model_name=None,
                      **kwargs):
  assert type(file_name) is str, "`file_name` must be name of .h5 file"
  import tensorflow.compat.v1 as tf
  tf.keras.backend.set_learning_phase(0)
  model = log.load_keras_model(
    file_name,
    custom_objects=custom_objects
  )
  return model_to_graph(
    log=log,
    model=model,
    model_name=model_name,
    **kwargs
    )

def model_file_to_graph(log, file_name, model_name=None):
  return log.model_h5_to_graph(file_name, model_name)

def model_to_graph(log, model, model_name=None, save_def=False):
  """
  quick and dirty converter from model to tf1 graph
  """
  log.P("Converting {} to tf1 graph".format(model.name))    
  model_name = model_name if model_name else model.name
  res = save_model_as_graph_to_models(
    log=log,
    model=model,
    model_name=model_name,
    save_def=save_def,
    )
  if res:
    return model.name + '.pb'
  else:
    return None


def save_model_as_graph_to_models(log, model, model_name,
                                  save_def=False):
  """
  """
  log._block_tf2()
  if model_name[-3:] != '.pb':
    model_name += '.pb'
  pb_file = os.path.join(log.get_models_folder(), model_name)
  try:
    save_model_as_graph(
      log=log,
      model=model,
      pb_file=pb_file,
      save_def=save_def,
      )    
    _res = True
  except:
    str_e = traceback.format_exc()
    log.P("ERROR: SaveModelAsGraphToModels failed: {}".format(str_e))
    _res = False
  return _res


def save_graph_to_models(log, session, output_tensor_list,
                         graph_name, input_names,
                         output_names=None,
                         ):
  log._block_tf2()
  if graph_name[-3:] != '.pb':
    graph_name += '.pb'
  pb_file = os.path.join(log.get_models_folder(), graph_name)
  fn = save_graph(log,
    session=session,
    output_tensor_list=output_tensor_list,
    pb_file=pb_file,
    check_input_names=input_names,
    )
  return fn


def save_model_as_graph(log, model, pb_file,
                        save_def=False):
  """
   saves keras model as frozen computational graph

  """
  import tensorflow.compat.v1 as tf
  # if tf.executing_eagerly():
  #   raise ValueError('`save_model_as_graph` from eager mode. Please start tf with eager execution disabled.')

  log._block_tf2() # seems is not actually required

  sess = tf.keras.backend.get_session()
  input_names = [x.name for x in model.inputs]
  out_tensors = model.outputs

  fn = save_graph(
    log=log,
    session=sess,
    output_tensor_list=out_tensors,
    pb_file=pb_file,
    check_input_names=input_names,
    save_def=save_def,
    )
  return fn


def save_graph_to_file(log, sess, graph, pb_file, input_tensor_names,
                       output_tensor_names):
  import tensorflow.compat.v1 as tf
  log._block_tf2()
  
  cfg_dict = OrderedDict()
  cfg_dict['DATE'] = dt.now().strftime("%Y%m%d_%H%M%S")
  # assume names are unique in graph
  assert len(output_tensor_names) != 0

  log.verbose_log("Saving graph with {} input(s) and {} output(s) ".format(
    len(input_tensor_names) if input_tensor_names else 0, len(output_tensor_names), ))

  # get and show inputs
  if input_tensor_names is not None:
    for i, name in enumerate(input_tensor_names):
      if name[-2:] != ':0':
        name += ':0'
      else:
        input_tensor_names[i] = input_tensor_names[i][:-2]
      cfg_dict['INPUT_{}'.format(i)] = name
      log.verbose_log("  Input: {}".format(name))

  # get and show outputs
  final_output_names = []
  for i, out_tensor in enumerate(output_tensor_names):
    out_name = out_tensor
    if out_tensor[-2:-1] != ':':
      out_tensor += ':0'
    cfg_dict['OUTPUT_{}'.format(i)] = out_tensor
    final_output_names.append(out_tensor[:-2])
    log.verbose_log("  Output: {}".format(final_output_names[-1] + ":0"))
  
  try:
    graph_def = graph.as_graph_def()
  except:
    graph_def = graph
  
  if int(tf.__version__.split('.')[1]) > 13 or log.is_tf2():
    constant_graph = tf.graph_util.convert_variables_to_constants(
      sess,
      graph_def,
      final_output_names)
  else:
    from tensorflow.python.framework import graph_util
    constant_graph = graph_util.convert_variables_to_constants(
      sess,
      graph_def,
      final_output_names
      )

 # prep path
  pb_file_path = log.get_models_folder()

  # save GraphDef to .pb
  from tensorflow.python.framework import graph_io
  log.verbose_log(" Saving {} in ...{}".format(pb_file, pb_file_path[-30:]))
  graph_io.write_graph(constant_graph, pb_file_path, pb_file,
                       as_text=False)
  log.save_models_json(cfg_dict, pb_file + '.txt')
  return

def save_graph(log, session, output_tensor_list,
               pb_file, check_input_names=None,
               save_def=False):
  """
  save TF computational graph
  session     : session where the tensors reside
  output_tensor_list : list of ouput tensors
  pb_file     : .pb filename
  check_input_names : list of input tensor names
  
  TODO: REFACTOR REQUIRED
  
  """
  import tensorflow.compat.v1 as tf
  
  log._block_tf2()
  
  klp = tf.keras.backend.eval(tf.keras.backend.learning_phase())
  if type(klp) != int or klp == 1:
    log.P("****************************************************************************", color='r')
    log.P("WARNING: you did not call `K.set_learning_phase(0)` before saving the graph.", color='r')
    log.P("         This graph is NOT optimzied for inference. Please set the learning ", color='r')
    log.P("         phase or use `model_h5_to_graph` method", color='r')
    log.P("****************************************************************************", color='r')

  tensor_list = output_tensor_list
  cfg_dict = OrderedDict()
  cfg_dict['DATE'] = dt.now().strftime("%Y%m%d_%H%M%S")
  cfg_dict['INPUTS'] = []
  cfg_dict['OUTPUTS'] = []
  # assume names are unique in graph
  assert len(tensor_list) != 0

  g = tensor_list[0].graph
  log.verbose_log("Saving graph with {} input(s) and {} output(s) ".format(
    len(check_input_names), len(tensor_list), ))

  # get and show inputs
  if check_input_names is not None:
    for i, name in enumerate(check_input_names):
      if name[-2:] != ':0':
        name += ':0'
      else:
        check_input_names[i] = check_input_names[i][:-2]
      cfg_dict['INPUTS'].append(name)
      log.verbose_log("  Input: {}".format(g.get_tensor_by_name(name)))

  # get and show outputs
  final_output_names = []
  for i, out_tensor in enumerate(tensor_list):
    out_name = out_tensor.name
    if out_name[-2:-1] != ':':
      out_name += ':0'
    cfg_dict['OUTPUTS'].append(out_name)
    final_output_names.append(out_name[:-2])
    log.verbose_log("  Output: {}".format(
      g.get_tensor_by_name(final_output_names[-1] + ":0")))

  # get graphdef
  graph_def = session.graph.as_graph_def()
  # now freeze the graph to constants
  if int(tf.__version__.split('.')[1]) > 13 or log.is_tf2():
    constant_graph = tf.graph_util.convert_variables_to_constants(
      session,
      graph_def,
      final_output_names)
  else:
    from tensorflow.python.framework import graph_util
    constant_graph = graph_util.convert_variables_to_constants(session,
                                                               graph_def,
                                                               final_output_names)
  # prep path
  pb_file_path, pb_file_name = os.path.split(pb_file)

  # save GraphDef to .pb
  from tensorflow.python.framework import graph_io
  log.verbose_log(" Saving {} in ...{}".format(pb_file_name, pb_file_path[-30:]))
  graph_io.write_graph(constant_graph, pb_file_path, pb_file_name,
                       as_text=False)
  log.save_models_json(cfg_dict, pb_file_name + '.txt')
  if save_def:
    from google.protobuf import json_format
    json_string = json_format.MessageToJson(graph_def)
    dct_def = json.loads(json_string)
    log.save_models_json(dct_def, pb_file_name + '.source.txt')

  return pb_file

def load_tf_graph_saved_model(log, pb_file, return_elements=None):
  import tensorflow.compat.v1 as tf
  from tensorflow.python.util import compat
  from tensorflow.python.platform import gfile
  from tensorflow.core.protobuf import saved_model_pb2
  log.verbose_log("Prep graph from [...{}]...".format(pb_file[-30:]))
  graph = None
  if os.path.isfile(pb_file):
    start_time = tm()      
    graph = tf.Graph()
    with graph.as_default():
      trt_graph_def = tf.GraphDef()
      with gfile.FastGFile(pb_file, 'rb') as f:
        data = compat.as_bytes(f.read())
        sm = saved_model_pb2.SavedModel()      
        sm.ParseFromString(data)
        trt_graph_def = sm.meta_graphs[0].graph_def
        tf.import_graph_def(
          trt_graph_def, 
          name='', #avoid 'import_' prefixes
          return_elements=return_elements
          )
    end_time = tm()
    log.verbose_log("Done preparing graph in {:.2f}s.".format(end_time - start_time))
  else:
    log.verbose_log(" FILE NOT FOUND [...{}]...".format(pb_file[-30:]))
  return graph

def load_tf_graph(log, pb_file, return_elements=None):
  import tensorflow.compat.v1 as tf
  log.verbose_log("Prep graph from [...{}]...".format(pb_file[-30:]))
  detection_graph = None
  if os.path.isfile(pb_file):
    start_time = tm()
    detection_graph = tf.Graph()
    with detection_graph.as_default():
      od_graph_def = tf.GraphDef()
      with tf.gfile.GFile(pb_file, 'rb') as fid:
        serialized_graph = fid.read()
        od_graph_def.ParseFromString(serialized_graph)
        tf.import_graph_def(od_graph_def, name='', return_elements=return_elements)
    end_time = tm()
    log.verbose_log("Done preparing graph in {:.2f}s.".format(end_time - start_time))
  else:
    log.verbose_log(" FILE NOT FOUND [...{}]...".format(pb_file[-30:]))
  return detection_graph

def load_graph_from_models(log, model_name, get_input_output=False):
  if model_name[-3:] != '.pb':
    model_name += '.pb'
  graph_file = os.path.join(log.get_models_folder(), model_name)
  tf_graph = load_tf_graph(log=log, pb_file=graph_file)
  if get_input_output:
    cfg = log.load_models_json(graph_file + '.txt')
    s_input = cfg['INPUT_0']
    s_output = cfg['OUTPUT_0']
    return tf_graph, s_input, s_output
  else:
    return tf_graph

def save_config_dict(log, model_name, tensor_names_dict):
  if model_name[-4:] != '.txt':
    model_name += '.txt'
  log.verbose_log("Saving cfg [{}] to models...".format(model_name))
  file_path = os.path.join(log.get_models_folder(), model_name)
  with open(file_path, 'w') as fp:
    json.dump(tensor_names_dict, fp, sort_keys=True, indent=4)
  return

def load_config_dict(log, model_name):
  if model_name[-4:] != '.txt':
    model_name += '.txt'
  log.verbose_log("Loading cfg [{}] from models...".format(model_name))
  file_path = os.path.join(log.get_models_folder(), model_name)
  with open(file_path, 'r') as fp:
    data = json.load(fp)
  return data

def save_tf2func_as_graph(log, func, 
                          tensorspec_kwargs, 
                          tensorspec_args=[],
                          folder='models', 
                          graph_name='frozen_graph.pb',
                          DEBUG=False):
  """
  The purpose of this function is to help you save a Tensorflow2 function into
  a frozen graph.
  Up until the moment of developing this function, no native Tensorflow method
  exists for this purpose.
  If you are in a spot to use a TF 'saved model' (custom, TFODAPI, etc) 
  and want to transform it to a frozen graph the following method should be 
  called with the parameters below.
  This method was inspired from https://leimao.github.io/blog/Save-Load-Inference-From-TF2-Frozen-Graph/
  
  Parameters
  ----------
  func : Tensorflow 2 function
    A function obtained by loading a 'saved model'.
  tensorspec_kwargs : dict
    'get_concrete_function' expects an kwargs in order to name the input graph tensors.
    tensorspec_kwargs should provide a key,value dictionary containing tensor descriptions.
    Ex: 
      tensorspec_kwargs={'image': tf.TensorSpec((None, None, None, 3), tf.float32)}
  tensorspec_args : args, optional
    'get_concrete_function' expects args in order to specify the output graph tensors.
    tensorspec_args should contain tensors descriptions.
    Ex: 
      tensorspec_args=[tf.TensorSpec((None, 100), tf.int32), tf.TensorSpec((None, 100, 4), tf.float32), tf.TensorSpec((None, 100), tf.float32)]
      The tensor descriptions above can describe some outputs tensors cooresponding
      to a object detection graph that outputs: classes, boxes, scores.
  folder : string, optional
    Path where the frozen graph will be saved. The default is 'models'.
  graph_name : string, optional
    Name of the frozen graph. The default is 'frozen_graph.pb'.
  DEBUG : bool, optional
    Used to print debug information. The default is False.

  Raises
  ------
  ValueError
    Raised when the folder destination path is not uderstood.

  Returns
  -------
  None.

  """
  import tensorflow as tf
  from tensorflow.python.framework.convert_to_constants import convert_variables_to_constants_v2
  
  if folder is None:
    folder = '.'
  elif folder in ['data', 'output', 'models']:
    folder = log.get_target_folder(target=folder)
  else:
    if not os.path.exists(folder):
      raise ValueError('Folder value not understood: {}'.format(folder))
  #endif
  
  
  # Get frozen ConcreteFunction
  func = func.get_concrete_function(*tensorspec_args, **tensorspec_kwargs)
  frozen_func = convert_variables_to_constants_v2(func)
  frozen_func.graph.as_graph_def()
  
  if DEBUG:
    tensors = log.get_tensors_in_tf_graph(frozen_func.graph)
    log.p("-" * 50)
    log.p("Frozen model tensors: ")
    for tensor in tensors:
      log.p(tensor)
  
    log.p("-" * 50)
    log.p("Frozen model inputs: ")
    log.p(frozen_func.inputs)
    log.p("Frozen model outputs: ")
    log.p(frozen_func.outputs)
  #endif
  
  # Save frozen graph from frozen ConcreteFunction to hard drive
  path = os.path.join(folder, graph_name)
  log.verbose_log('Frozen graph will be saved into: {}'.format(path))
  tf.io.write_graph(
    graph_or_graph_def=frozen_func.graph,
    logdir=folder,
    name=graph_name,
    as_text=False
    )
  log.p('The frozen graph was saved successfully in ...{}'.format(path[-50:]))
  return


def tfodapi2_ckpt_to_graph(log, 
                           path_pipeline_config, 
                           path_checkpoint, 
                           tensorspec_kwargs,
                           tensorspec_args=[],
                           folder='models',
                           graph_name='frozen_graph.pb',
                           DEBUG=False
                          ):
  """
  The purpose of this method is to help you easily load a TFODAPI2 checkpoint
  and saved it into a frozen graph.

  Parameters
  ----------
  path_pipeline_config : string
    Path to model pipeline config.
  path_checkpoint : string
    Path to model checkpoint.
  for the rest of the parameters please see save_tf2func_as_graph method
  Returns
  -------
  None

  """
  import tensorflow as tf
  from object_detection.utils import config_util
  from object_detection.builders import model_builder
  
  def get_model_detection_function(model):
    """Get a tf.function for detection."""
    @tf.function
    def detect_fn(image):
      """Detect objects in image."""
      image, shapes = model.preprocess(image)
      prediction_dict = model.predict(image, shapes)
      detections = model.postprocess(prediction_dict, shapes)
      return detections, prediction_dict, shapes
    #enddef
    return detect_fn
  #enddef
  
  assert os.path.exists(path_pipeline_config)
  assert os.path.split(os.path.split(path_checkpoint)[0])
  
  # Load pipeline config and build a detection model
  configs = config_util.get_configs_from_pipeline_file(path_pipeline_config)
  model_config = configs['model']
  detection_model = model_builder.build(
    model_config=model_config, 
    is_training=False
    )
  
  # Restore checkpoint
  ckpt = tf.compat.v2.train.Checkpoint(model=detection_model)
  ckpt.restore(path_checkpoint).expect_partial()

  detect_fn = get_model_detection_function(detection_model)
  
  if DEBUG:
    img = np.random.randint(
      low=0, 
      high=255, 
      size=(2, 300, 300, 3)
      ).astype(np.float32)
    detections, prediction_dict, shapes = detect_fn(img)
    log.p('Shapes: {}'.format(shapes.numpy()))
  #endif
  
  save_tf2func_as_graph(log,
    func=detect_fn,    
    tensorspec_kwargs=tensorspec_kwargs,
    tensorspec_args=tensorspec_args,
    folder=folder,
    graph_name=graph_name,
    DEBUG=DEBUG
    )
  return

###
### end keras/tf converter
###

### TF1 Graph helpers

def combine_graphs_tf1(log, lst_graphs, lst_names):
  """
  will return a graph that combines all given graphs
  individual tensors of graph `i` in `lst_graphs` can be accessed via
  `lst_names[i] + '/TENSOR_NAME'.
  """
  import tensorflow.compat.v1 as tf
  assert len(lst_graphs) == len(lst_names)
  gdefs = []
  for graph in lst_graphs:
    gdefs.append(graph.as_graph_def())
  full_graph = tf.Graph()
  log.P("Creating one graph out of {} graphs: {}".format(len(lst_graphs), lst_names))
  with full_graph.as_default():
    for gdef, gname in zip(gdefs, lst_names):
      log.P("  Adding {}".format(gname))
      tf.import_graph_def(graph_def=gdef, name=gname)
  return full_graph

### END TF1 Graph helpers

###############################
###                         ###
###  Ensemble models area   ###
###                         ###
###############################

def _gen_ensemble(log, models, model_name):
  import tensorflow as tf
  tf_inp = tf.keras.layers.Input(models[0].input.shape[1:], name='ensemble_input')
  lst_tensors = [x(tf_inp) for x in models]
  tf_out = tf.keras.layers.average(lst_tensors)
  model_ensemble = tf.keras.models.Model(inputs=tf_inp, outputs=tf_out, name=model_name)
  return model_ensemble

def ensemble_generate_tf1(log, lst_files_from_models, model_name):
  return ensemble_generate(log, lst_files_from_models=lst_files_from_models,
                                model_name=model_name,
                                generate_tf1_graph=True)

def ensemble_generate(log, lst_files_from_models, model_name, generate_tf1_graph=False):
  """
  takes as input a list of model files and generates a average-based ensemble model or tf1 graph
  assumes all inputs are identical
  """
  if generate_tf1_graph:
    log._block_tf2()

  log.P("Generating averaged ensemble tf1x graph from {} files.".format(len(lst_files_from_models)))
  lst_models = [log.load_keras_model(x) for x in lst_files_from_models]
  for i in range(len(lst_models)):
    lst_models[i]._name = lst_models[i]._name + str(i)
    
  model_ensemble = _gen_ensemble(log, models=lst_models, model_name=model_name)

  if generate_tf1_graph:
    _res = save_model_as_graph_to_models(log, model_ensemble, model_name)
                                              # output_layers=[model_ensemble.layers[-1]],
                                              # input_layers=[model_ensemble.layers[0]])
  else:
    _res = log.save_keras_model(model_ensemble, label=model_name)

  _res = model_name if _res else None
  return _res

def ensemble_search_tf1(log, test_func,
                        X_test, y_test,
                        nr_cand=3, name='ens',
                        lst_cand_files=None,
                        exlude_fn='ens',
                        test_func_extra_param=None,
                        batch_size=32,
                        max_iters=None,
                        ):
  return ensemble_search(log, test_func=test_func,
                              X_test=X_test, y_test=y_test,
                              nr_cand=nr_cand, name=name,
                              lst_cand_files=lst_cand_files,
                              exlude_fn=exlude_fn,
                              test_func_extra_param=test_func_extra_param,
                              batch_size=batch_size,
                              max_iters=max_iters,
                              generate_tf1_graph=True
                              )

def ensemble_search(log, test_func,
                    X_test, y_test,
                    nr_cand=3,
                    lst_cand_files=None,
                    name='ens',
                    exlude_fn='ens',
                    test_func_extra_param=None,
                    batch_size=32,
                    max_iters=None,
                    generate_tf1_graph=False,
                    output_csv_name=None
                    ):
  """
  generates all models or tf1 graph ensemble candidates out of `_models` folder bassed on `test_func`
  that has signature either f(preds, y_test) or f(preds, y_test, test_func_extra_param)
  and returns a performance indicator value such as `dev_acc`

  test_func : f(preds, y_test, test_func_extra_param=None) calculates the test results based on predictions
  X_test : input for models
  y_test : gold
  nr_cand: number of candidates for each ensemble
  max_iters : limit to a certain number of combinations
  generate_tf1_graph : will output a tf1 graph instead of a model
  output_csv_name: will save results into specified file
  """

  from itertools import combinations
  import pandas as pd
  from time import time
  import tensorflow as tf
  if generate_tf1_graph:
    import tensorflow.compat.v1 as tf
    log._block_tf2()

  if lst_cand_files:
    lst_all_cand_files = lst_cand_files
  else:
    lst_all_cand_files = [x for x in os.listdir(log.get_models_folder()) if x.endswith('.h5') and exlude_fn not in x]
  # endif
  lst_cands = list(combinations(lst_all_cand_files, r=nr_cand))
  n_files = len(lst_all_cand_files)
  nr_candidates = len(lst_cands)
  log.P("Testing {} base .pb graph models for a total of {} ensembles".format(n_files, nr_candidates))
  dct_results = {'MODEL': [], 'TEST': [], 'TIME': [], 'SZ_MB': [], 'COMPONENTS': []}

  t_iter = np.inf
  t_iter_models = np.inf

  log.P("Testing base candidates...")
  for i_sgle_cand, fn_model in enumerate(lst_all_cand_files):
    log.start_timer('MODELS_FULL_ITER')
    t_left = (n_files - i_sgle_cand) * t_iter_models
    log.P("Testing {:.1f}% done. Remaining time: {:.1f} hrs.".format(
      (i_sgle_cand + 1) / n_files * 100, t_left / 3600))
    tf.keras.backend.clear_session()
    model = log.load_keras_model(fn_model)
    for _ in range(2):
      t_start = time()
      np_preds = model.predict(X_test)
      t_end = time()
    t_run_time = t_end - t_start
    log.P("  Output preds {}".format(np_preds.shape))

    if test_func_extra_param is not None:
      test_result = test_func(preds=np_preds, y_test=y_test, test_func_extra_param=test_func_extra_param)
    else:
      test_result = test_func(preds=np_preds, y_test=y_test)

    dct_results['MODEL'].append(fn_model)
    dct_results['TEST'].append(test_result)
    dct_results['COMPONENTS'].append(fn_model)
    dct_results['TIME'].append(t_run_time)
    fn_full_h5 = log.get_models_file(fn_model)
    dct_results['SZ_MB'].append(round(os.path.getsize(fn_full_h5) / (1024 ** 2), 3))
    log.end_timer('MODELS_FULL_ITER')
    t_iter_models = log.get_timer_mean('MODELS_FULL_ITER')

  if max_iters is not None:
    idxs = np.random.choice(nr_candidates, size=max_iters, replace=False)
  else:
    idxs = np.arange(nr_candidates)

  n_tries = idxs.shape[0]
  for i_cand, idx in enumerate(idxs):
    lst_files = lst_cands[idx]
    log.start_timer('ENSEMBLE_FULL_ITER')
    t_left = (n_tries - i_cand) * t_iter
    log.P("Testing {:.1f}% done. Remaining time: {:.1f} hrs.".format(
      (i_cand + 1) / n_tries * 100, t_left / 3600))
    tf.keras.backend.clear_session()
    model_name = '{}_{:04d}'.format(name, i_cand + 1)
    fn_ens = ensemble_generate(log, 
                               lst_files,
                               model_name=model_name,
                               generate_tf1_graph=generate_tf1_graph)
    if fn_ens is None:
      log.raise_error("Could not load {}".format(fn_ens))

    tf.keras.backend.clear_session()
    if generate_tf1_graph:
      tf_graph = load_graph_from_models(log, fn_ens)
      tf_sess = tf.Session(graph=tf_graph)
      dct_config = log.load_models_json(fn_ens + '.pb.txt')
      tf_input = tf_graph.get_tensor_by_name(dct_config['INPUT_0'])
      tf_output = tf_graph.get_tensor_by_name(dct_config['OUTPUT_0'])
      # now test each graph twice and keep second set of results
    else:
      ens_model = log.load_keras_model(fn_ens)

    for _ in range(2):
      t_start = time()
      if generate_tf1_graph:
        nr_batches = X_test.shape[0] // batch_size
        lst_preds = []
        tf_runoptions = tf.RunOptions(report_tensor_allocations_upon_oom=True)
        for nr_batch in range(nr_batches + 1):
          start_idx = nr_batch * batch_size
          stop_idx = (nr_batch + 1) * batch_size
          x_batch = X_test[start_idx:stop_idx]
          probs = tf_sess.run(tf_output,
                                  feed_dict={tf_input: x_batch},
                                  options=tf_runoptions)
          lst_preds.append(probs)
      else:
        np_preds = ens_model.predict(X_test)
      # endfor 2
      t_end = time()
    t_run_time = t_end - t_start

    if generate_tf1_graph:
      np_preds = np.concatenate(lst_preds, axis=0)

    log.P("  Output preds {}".format(np_preds.shape))

    if test_func_extra_param is not None:
      test_result = test_func(preds=np_preds, y_test=y_test, test_func_extra_param=test_func_extra_param)
    else:
      test_result = test_func(preds=np_preds, y_test=y_test)

    dct_results['MODEL'].append(model_name)
    dct_results['TEST'].append(test_result)
    dct_results['COMPONENTS'].append([_fn[-15:] for _fn in lst_files])
    dct_results['TIME'].append(t_run_time)
    

    if generate_tf1_graph:
      fn_full_pb = log.get_model_file(fn_ens + '.pb')
      fn_full_txt = log.get_model_file(fn_ens + '.pb.txt')
      dct_results['SZ_MB'].append(round(os.path.getsize(fn_full_pb) / (1024 ** 2), 3))
      fn_new_pb = os.path.join(log.get_models_folder(), fn_ens + "_T_{:.3f}.pb".format(test_result))
      fn_new_txt = os.path.join(log.get_models_folder(), fn_ens + "_T_{:.3f}.pb.txt".format(test_result))
      os.rename(fn_full_pb, fn_new_pb)
      os.rename(fn_full_txt, fn_new_txt)
    else:
      fn_full_mdl = log.get_model_file(fn_ens + '.h5')
      dct_results['SZ_MB'].append(round(os.path.getsize(fn_full_mdl) / (1024 ** 2), 3))
      fn_new_mdl = os.path.join(log.get_models_folder(), fn_ens + "_T_{:.3f}.h5".format(test_result))
      os.rename(fn_full_mdl, fn_new_mdl)

    log.end_timer('ENSEMBLE_FULL_ITER')
    t_iter = log.get_timer_mean('ENSEMBLE_FULL_ITER')
    df = pd.DataFrame(dct_results).sort_values('TEST')
    
    log.set_nice_prints(df_precision=3, precision=3)
    log.P("Results so far based on {} trials:\n{}".format(i_cand + 1, df))
  # end ensemble search loop
  if output_csv_name:
    df.to_csv(os.path.join(log.get_output_folder(), output_csv_name))
  return df

###
### END Ensemble models area
###





import os
import sys
sys.path.append(os.getcwd())

import argparse
import json

from naeural_core import Logger
from naeural_core.local_libraries.model_server_v2 import FlaskModelServer

if __name__ == '__main__':
  parser = argparse.ArgumentParser()

  parser.add_argument(
    '-b', '--base_folder',
    type=str, default='libraries',
    help='Logger base folder'
  )

  parser.add_argument(
    '-a', '--app_folder',
    type=str, default='_logger_cache',
    help='Logger app folder'
  )

  parser.add_argument(
    '--config_endpoint', type=json.loads,
    help='JSON configuration of the endpoint'
  )

  parser.add_argument(
    '--host', type=str, default='127.0.0.1'
  )

  parser.add_argument(
    '--port', type=int, default=5002
  )

  parser.add_argument(
    '--execution_path', type=str, default='/analyze'
  )

  parser.add_argument(
    '--workers_location', type=str
  )

  parser.add_argument(
    '--worker_name', type=str
  )


  parser.add_argument(
    '--microservice_name', type=str
  )

  parser.add_argument(
    '--microservice_code', type=str
  )

  parser.add_argument(
    '--worker_suffix', type=str, default='Worker'
  )

  parser.add_argument(
    '--nr_workers', type=int
  )

  parser.add_argument(
    '--use_tf', action='store_true'
  )

  args = parser.parse_args()
  base_folder = args.base_folder
  app_folder = args.app_folder
  config_endpoint = args.config_endpoint
  host = args.host
  port = args.port
  execution_path = args.execution_path
  workers_location = args.workers_location
  worker_name = args.worker_name
  worker_suffix = args.worker_suffix
  microservice_name = args.microservice_name
  microservice_code = args.microservice_code
  nr_workers = args.nr_workers
  use_tf = args.use_tf
  
  if microservice_code is None:
    microservice_code = Logger.name_abbreviation(microservice_name)

  log = Logger(
    lib_name=microservice_code,
    base_folder=base_folder,
    app_folder=app_folder,
    TF_KERAS=False, # use_tf
    max_lines=3000
  )

  svr = FlaskModelServer(
    log=log,
    workers_location=workers_location,
    worker_name=worker_name,
    microservice_name=microservice_name,
    worker_suffix=worker_suffix,
    host=host,
    port=port,
    config_endpoint=config_endpoint,
    execution_path=execution_path,
    nr_workers=nr_workers
  )
